/*****************************************************************************/
#include <iostream.h>
#include <fstream.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "air-hockey.h"
#include "../procedures/procedures.hpp"
#include "../idle/idle.hpp"

/*****************************************************************************/

/*****************************************************************************/

typedef struct hit
{
  /* lookup variables */
  float x;
  float y;
  float vx;
  float vy;

  /* hit information */
  int type;
  
  float hit_x;
  float hit_y;
  float hit_t;
  float hit_d;
  float hit_v;
  float target_x;
  float target_y;

} Hit;

/*****************************************************************************/

void
do_dynamics_and_handle_collisions( float seconds, float mouse_x, float mouse_y,
				  float paddle2_x, float paddle2_y, Game *g );
void
load_communication_buffer( Game *g, Communication_Buffer *cb );

void
generate_paddle2_motion_x( Game *g, float *paddle2_target );

void
generate_paddle2_motion_xy( Game *g, float *paddle2_target );

void generate_paddle2_motion_imitate( Game *g, float *paddle2_target );
void generate_paddle2_motion_simple( Game *g, float *paddle2_target );
void generate_paddle2_motion_target( Game *g, float *paddle2_target );

void write_samples();

Hit *lookup( float x, float y, float vx, float vy );

void plan_hit( Game *g, Hit *h, float *paddle2_desireds, int first_time );

float normalize_vector( float *vector, int length );
void read_hits( char * );

void dcb_plan_hit( Game *g, Hit *h, float *paddle2_desireds, int first_time );
void dcb_procedure_hit(Game*g,Hit *h, float *paddle2_desireds, int first_time );
void dcbf_plan_hit( Game *g, Hit *h, float *paddle2_desireds, int first_time );
/*****************************************************************************/

Game the_game, *game;

Sample *sample_ptr;
Sample samples[NPOINTS];

#define N_DEBUG_POINTS 400
#define N_DEBUG_DIMENSIONS 20
float debug_array[N_DEBUG_POINTS][N_DEBUG_DIMENSIONS];
int debug_index = 0;

#define DEBUG_DYN 11

static float theta, theta_des;
static float theta_hit;
float PVx,PVy,OVx,OVy,IVx,IVy;
float max_paddle_velocity=3.0;  // (meters/sec) The maximum velocity of the cyberhuman's paddle.

float PBx, PBy;

Procedures procedures;
Idle idle;
Communication_Buffer * g_cb; //Pointer to the communications buffer.
/*****************************************************************************/
/*NDEBUG,WIN32,_WINDOWS,NO_STRICT */
/*****************************************************************************/
/* Global variables */
int use_nn =1;

void
init_game( Communication_Buffer *cb )
{
  Game *g;
  int i, j;
  g_cb = cb;
  
  procedures.LoadProcedures("procedures");
  idle.LoadIdlePoints("idle.dbs");

  g = game = &the_game;

  /*Ensure the player can move if max_reach is not initialized. */
  if (cb->max_reach <= 0)
	cb->max_reach=board_length_div2;

  /* g->max_reach=(board_length_div2.0)-(cb->max_reach);*/
  g->max_reach=(cb->max_reach);

  g->ball_radius = BALL_RADIUS;
  /* g->ball_radius = 0.2; */
  g->paddle1_radius = PADDLE_RADIUS;
  g->paddle2_radius = PADDLE_RADIUS;

  g->bp_threshold = (g->ball_radius + g->paddle1_radius)*
	(g->ball_radius + g->paddle1_radius);

  g->left = -board_width_div2;
  g->right = board_width_div2;
  g->bottom = -board_length_div2;
  g->top = board_length_div2;

  g->goal1_left = goal1_left;
  g->goal1_right = goal1_right;

  g->goal2_left = goal1_left;
  g->goal2_right = goal1_right;

  g->in_goal1 = 0;
  g->score1 = 0;

  g->in_goal2 = 0;
  g->score2 = 0;

  g->restitution = restitution; //Set in air-hockey.h
  g->friction = friction; //Set in air-hockey.h

  g->time = 0.0;
  g->last_time = 0.0;

  g->status_bits = 0;

  g->ball[XX] = -0.1f;
  g->ball[YY] = -0.1f;
  g->ball[ZZ] = 0.0f;

  g->ball_velocity[XX] = 0.7f;
  g->ball_velocity[YY] = 1.4f;
  g->ball_velocity[ZZ] = 0.0f;

  g->paddle1[XX] = 0.0f;
  g->paddle1[YY] = g->bottom/2.0;
  g->paddle1[ZZ] = 0.0;

  g->paddle1_velocity[XX] = 0.0;
  g->paddle1_velocity[YY] = 0.0;
  g->paddle1_velocity[ZZ] = 0.0;

  g->paddle2[XX] = 0.0;
  g->paddle2[YY] = g->top/2;
  g->paddle2[ZZ] = 0.0;

  g->paddle1_velocity[XX] = 0.0;
  g->paddle1_velocity[YY] = 0.0;
  g->paddle1_velocity[ZZ] = 0.0;

  g->paddle2_velocity[XX] = 0.0;
  g->paddle2_velocity[YY] = 0.0;
  g->paddle2_velocity[ZZ] = 0.0;

  sample_ptr = samples;

  load_communication_buffer( g, cb );

  read_hits( "hits" );

  for( i = 0; i < N_DEBUG_POINTS; i++ )
	for( j = 0; j < N_DEBUG_DIMENSIONS; j++ )
	  debug_array[i][j] = 0.0;
}

/*****************************************************************************/

void
load_communication_buffer( Game *g, Communication_Buffer *cb )
{
  cb->left = g->left;
  cb->right = g->right;
  cb->top = g->top;
  cb->bottom = g->bottom;

  cb->goal1_left = g->goal1_left;
  cb->goal1_right = g->goal1_right;

  cb->goal2_left = g->goal2_left;
  cb->goal2_right = g->goal2_right;

  cb->ball_radius = g->ball_radius;
  cb->paddle1_radius = g->paddle1_radius;
  cb->paddle2_radius = g->paddle2_radius;

  cb->ball[0] = g->ball[0];
  cb->ball[1] = g->ball[1];
  cb->ball[2] = 0.0;
  cb->paddle1[0] = g->paddle1[0];
  cb->paddle1[1] = g->paddle1[1];
  cb->paddle1[2] = 0.0;
  cb->paddle2[0] = g->paddle2[0];
  cb->paddle2[1] = g->paddle2[1];
  cb->paddle2[2] = 0.0;

  cb->score1 = g->score1;
  cb->score2 = g->score2;

  cb->status_bits = g->status_bits;
}

/*****************************************************************************/

int count = 0;
float first_call_time = 0.0;

//
// Reset the the Hockey board.  Put the puck on one side and have it move
// slowly to the other side.  This will ensure the cyber player's algorithm
// is triggered.  It will move in a random x direction.
//
void
resetHockeyBoard(int side) 
{

	static int firsttime = 1;
	float by,bvy, bvx=0;

	if(firsttime)
		srand( (unsigned)time( NULL ) );
	bvx = (float) 0.07 - (float) rand()/RAND_MAX* 0.14;
	//cout << "XVEL:" <<bvx<<endl;
	if(side == 1) { //Move toward cyber player. 
		by=-0.05f; 
		bvy=0.12f;
	} else { // Move toward human.
		by=0.05f;
		bvy=-0.12f;
	}
	game->ball[XX] = 0;
	game->ball[YY] = by;
	if(!firsttime) {
		game->ball_velocity[XX] = bvx;
		game->ball_velocity[YY] = bvy;
	} else {
		firsttime=0;
	}
	// set the status bits
	game->status_bits = RESET_PRESSED;
}


void
simulate_up_to_time( float stop_time, float mouse_x, float mouse_y,
			Communication_Buffer *cb )
{
	Game *g;
	float time, time_increment;
	float mouse[2];
	float mouse_increment[2];
	float paddle2[2];
	//  float paddle2_target[2];
	float paddle2_increment[2];
	float paddle2_desireds[4], time_interval, paddle2_dist, max_paddle2_dist;

	if ( count == 0 )
		first_call_time = stop_time;
	count++;

	g = game;

	//generate_paddle2_motion_xy( g, paddle2_desireds );
	generate_paddle2_motion_imitate( g, paddle2_desireds );
	//generate_paddle2_motion_target( g, paddle2_desireds );

	// These two lines will keep the paddle in one location.
	//paddle2_desireds[XX]=0.22;
	//paddle2_desireds[YY]=0.46;

	time_interval = stop_time - g->time;  // The total time for this interval.

	//Ensure paddle two does not exceed the velocity limit.
	paddle2_dist = _hypot((g->paddle2[XX]-paddle2_desireds[XX]),
							(g->paddle2[YY]-paddle2_desireds[YY]));
	max_paddle2_dist= max_paddle_velocity*time_interval;
	if (paddle2_dist > max_paddle2_dist) {
		// The paddle is moving too fast, reset the desired location.
		float percent_dist = max_paddle2_dist/paddle2_dist;
		paddle2_desireds[XX] = percent_dist*(paddle2_desireds[XX]-g->paddle2[XX])+g->paddle2[XX];
		paddle2_desireds[YY] = percent_dist*(paddle2_desireds[YY]-g->paddle2[YY])+g->paddle2[YY];
		//cout<<"Adjusting....."<<percent_dist<<endl;
	}

  g->status_bits = 	game->status_bits;
  g_cb->track_ball->Simulate(g->time); 
  if(g_cb->track_ball == NULL)
	cerr << "No Track Ball"<<endl;

  // Ensure paddle one stays on the board.
  if ( mouse_x < g->left + g->paddle1_radius )
	mouse_x = g->left + g->paddle1_radius;
  if ( mouse_x > g->right - g->paddle1_radius )
	mouse_x = g->right - g->paddle1_radius;
  if ( mouse_y < g->bottom + g->paddle1_radius )
	mouse_y = g->bottom + g->paddle1_radius;
  if ( mouse_y > g->top - g->paddle1_radius )
	mouse_y = g->top - g->paddle1_radius;

  // Ensure paddle 2 stays on the board.
  if ( paddle2_desireds[XX] < g->left + g->paddle2_radius )
	paddle2_desireds[XX] = g->left + g->paddle2_radius;
  if ( paddle2_desireds[XX] > g->right - g->paddle2_radius )
	paddle2_desireds[XX] = g->right - g->paddle2_radius;
  if ( paddle2_desireds[YY] < g->bottom + g->paddle2_radius )
	paddle2_desireds[YY] = g->bottom + g->paddle2_radius;
  if ( paddle2_desireds[YY] > g->top - g->paddle2_radius )
	paddle2_desireds[YY] = g->top - g->paddle2_radius;

  /* break down big step into NITER smaller steps */
  time_increment = time_interval/NITER;
  if(time_increment<0.001)	{ //Do not simulate intervals less than 1msec.
	  //cerr <<"======== Too small ==========="<<endl;
	  return;
  }
  mouse_increment[XX] = (mouse_x - g->paddle1[XX])/NITER;
  mouse_increment[YY] = (mouse_y - g->paddle1[YY])/NITER;

  paddle2_increment[XX] = (paddle2_desireds[XX]-g->paddle2[XX])/NITER;
  paddle2_increment[YY] = (paddle2_desireds[YY]-g->paddle2[YY])/NITER;

  g->paddle1_velocity[XX]=(mouse_x-g->paddle1[XX])/
	(stop_time - g->time);
  g->paddle1_velocity[YY]=(mouse_y-g->paddle1[YY])/
	(stop_time - g->time);
  g->paddle2_velocity[XX]=(paddle2_desireds[XX]-g->paddle2[XX])/
	(stop_time - g->time);
  g->paddle2_velocity[YY]=(paddle2_desireds[YY]-g->paddle2[YY])/
	(stop_time - g->time);

  g->score1 = cb->score1;
  g->score2 = cb->score2;

  for( time = g->time + time_increment,
	 mouse[XX] = g->paddle1[XX] + mouse_increment[XX],
	 mouse[YY] = g->paddle1[YY] + mouse_increment[YY],
	 paddle2[XX] = g->paddle2[XX] + paddle2_increment[XX],
	 paddle2[YY] = g->paddle2[YY] + paddle2_increment[YY]
	 ;
	   time < stop_time + 0.0001
	 ;
	   time += time_increment,
	 mouse[XX] += mouse_increment[XX],
	 mouse[YY] += mouse_increment[YY],
	 paddle2[XX] += paddle2_increment[XX],
	 paddle2[YY] += paddle2_increment[YY]
	   ) {
	do_dynamics_and_handle_collisions( time, mouse[XX], mouse[YY],
					   paddle2[XX], paddle2[YY], g );

  }

	   //Load the communications buffer.
  cb->ball[0] = g->ball[XX];
  cb->ball[1] = g->ball[YY];
  cb->paddle1[0] = g->paddle1[XX];
  cb->paddle1[1] = g->paddle1[YY];
  cb->paddle2[0] = g->paddle2[XX];
  cb->paddle2[1] = g->paddle2[YY];
  cb->status_bits = g->status_bits;
  cb->score1 = g->score1;
  cb->score2 = g->score2;

  if ( sample_ptr - samples < NPOINTS ) {
	sample_ptr->status_bits = g->status_bits;
	sample_ptr->time = stop_time;
	sample_ptr->ball[0] = g->ball[0];
	sample_ptr->ball[1] = g->ball[1];
	sample_ptr->paddle1[0] = g->paddle1[0];
	sample_ptr->paddle1[1] = g->paddle1[1];
	sample_ptr->paddle2[0] = g->paddle2[0];
	sample_ptr->paddle2[1] = g->paddle2[1];
	sample_ptr++;
  } else if ( sample_ptr - samples == NPOINTS ){
	write_samples();
	sample_ptr = samples;
  }

  if ( (count%1000) == 0 )
	printf( "%g elapsed time; %g seconds/cycle\n", stop_time - first_call_time,
	(stop_time - first_call_time)/count );
}

/*****************************************************************************/
/*
 * This is used for predicting the ball characteristics for some time
 * in the future.
 *
 */
void
do_dynamics_collisions(
			   float seconds, //Time to iterate over.
			   float *o_pos, //ball position.
			   float *o_vel, //ball velocity.
			   float *n_pos, //new ball position.
			   float *n_vel //new ball velocity. 
			   )
{
  int i;
  float new_v;
  int in_goal1 = FALSE;
  int in_goal2 = FALSE;
  Game * g = game;



  for( i = 0; i < 2; i++ )	{
	new_v =  o_vel[i]*(1.0 - g->friction*seconds);
	n_pos[i] =o_pos[i] + 0.5*(o_vel[i] + new_v)*seconds;
	n_vel[i] = new_v;
  }

  /* Handle ball wall collisions. */
  if ( n_pos[XX] < g->left + g->ball_radius ) {
	n_pos[XX] = 2*(g->left + g->ball_radius) - n_pos[XX];
	n_vel[XX] = -g->restitution*n_vel[XX];
	//g->status_bits |= LEFT_CONTACT;
	//printf("++++ LC:%0.3f\n", n_vel[XX]);
  }
  if ( n_pos[XX] > g->right - g->ball_radius ) {
	n_pos[XX] = 2*(g->right - g->ball_radius) - n_pos[XX];
	n_vel[XX] = -g->restitution*n_vel[XX];
	//g->status_bits |= RIGHT_CONTACT;
  }
  if ( n_pos[YY] < g->bottom + g->ball_radius ) {
	n_pos[YY] = 2*(g->bottom + g->ball_radius) - n_pos[YY];
	n_vel[YY] = -g->restitution*n_vel[YY];
	//g->status_bits |= BOTTOM_CONTACT;
	if ( ( g->goal2_left <= g->ball[XX] ) &&
	 ( g->goal2_right >= g->ball[XX] ) )
	  in_goal2 = TRUE;
  }
  if ( n_pos[YY] > g->top - g->ball_radius ) {
	n_pos[YY] = 2*(g->top - g->ball_radius) - n_pos[YY];
	n_vel[YY] = -g->restitution*n_vel[YY];
	//->status_bits |= TOP_CONTACT;
	if ( ( g->goal1_left <= g->ball[XX] ) &&
	 ( g->goal1_right >= g->ball[XX] ) )
	  in_goal1 = TRUE;
  }
  if(in_goal1==in_goal2) return; //used to eliminate warning.
}

/*****************************************************************************/
void
do_dynamics_and_handle_collisions( float seconds, float mouse_x, float mouse_y,
				  float paddle2_x, float paddle2_y, Game *g )
{
	int i;
	float ball_paddle_distance;
	float c, s;
	float v_rotated_x, v_rotated_y;
	float vx, vy;
	float new_v;
	int in_goal1 = FALSE;
	int in_goal2 = FALSE;

	g->last_time = g->time;
	g->time = seconds;

	g->paddle1[XX] = mouse_x;
	g->paddle1[YY] = mouse_y;

	g->paddle2[XX] = paddle2_x;
	g->paddle2[YY] = paddle2_y;

	for( i = 0; i < ND; i++ ) {
		new_v = g->ball_velocity[i]*(1.0 - g->friction*(g->time - g->last_time));
		g->ball[i] += 
			0.5*(g->ball_velocity[i] + new_v)*(g->time - g->last_time);
		g->ball_velocity[i] = new_v;
	}

	/* Handle ball wall collisions. */
	if ( g->ball[XX] < g->left + g->ball_radius ) {
		g->ball[XX] = 2*(g->left + g->ball_radius) - g->ball[XX];
		g->ball_velocity[XX] = -g->restitution*g->ball_velocity[XX];
		g->status_bits |= LEFT_CONTACT;
	}
	if ( g->ball[XX] > g->right - g->ball_radius ) {
		g->ball[XX] = 2*(g->right - g->ball_radius) - g->ball[XX];
		g->ball_velocity[XX] = -g->restitution*g->ball_velocity[XX];
		g->status_bits |= RIGHT_CONTACT;
	}
	if ( g->ball[YY] < g->bottom + g->ball_radius ) {	
		g->ball[YY] = 2*(g->bottom + g->ball_radius) - g->ball[YY];
		g->ball_velocity[YY] = -g->restitution*g->ball_velocity[YY];
		g->status_bits |= BOTTOM_CONTACT;
		if ( ( g->goal2_left <= g->ball[XX] ) &&
			( g->goal2_right >= g->ball[XX] ) )
			in_goal2 = TRUE;
	}
	if ( g->ball[YY] > g->top - g->ball_radius ) {
		g->ball[YY] = 2*(g->top - g->ball_radius) - g->ball[YY];
		g->ball_velocity[YY] = -g->restitution*g->ball_velocity[YY];
		g->status_bits |= TOP_CONTACT;
		if ( ( g->goal1_left <= g->ball[XX] ) &&
			( g->goal1_right >= g->ball[XX] ) )
			in_goal1 = TRUE;
	}

	if ( ( in_goal1 && !(g->in_goal1) ) ||( in_goal2 && !(g->in_goal2) ) ) {
		if ( in_goal1 && !(g->in_goal1) ) {
			g->score1++;
			g->status_bits |= PLAYER_ONE_SCORED;
		}
		if ( in_goal2 && !(g->in_goal2) ) {
			g->score2++;
		g->status_bits |= PLAYER_TWO_SCORED;
		}
		printf( "%d %d\n", g->score1, g->score2 );
	}
	g->in_goal1 = in_goal1;
	g->in_goal2 = in_goal2;

  /* Handle ball paddle1 collisions. */
  ball_paddle_distance = 0.0;
  for ( i = 0; i < 2; i++ )
	ball_paddle_distance +=
	  (g->ball[i] - g->paddle1[i])*(g->ball[i] - g->paddle1[i]);

  if ( ball_paddle_distance < g->bp_threshold )
	 {
	   g->status_bits |= PADDLE_ONE_CONTACT;

	   /* Okay, rotate everything so collision is horizontal from right. */
	   ball_paddle_distance = sqrt( (double) ball_paddle_distance );
	   /* cosine angle */
	   c = (g->ball[XX] - g->paddle1[XX])/ball_paddle_distance; 
	   /* sin angle */
	   s = (g->ball[YY] - g->paddle1[YY])/ball_paddle_distance; 

	   vx = g->ball_velocity[XX] - g->paddle1_velocity[XX];
	   vy = g->ball_velocity[YY] - g->paddle1_velocity[YY];

	   v_rotated_x = c*vx + s*vy;
	   v_rotated_y = -s*vx + c*vy;

	   /* collision flips velocity */
	   if ( v_rotated_x < 0.0 ) {
	 v_rotated_x = -g->restitution*v_rotated_x;
	   }

	   vx = c*v_rotated_x - s*v_rotated_y;
	   vy = s*v_rotated_x + c*v_rotated_y;
	   
	   g->ball_velocity[XX] = vx + g->paddle1_velocity[XX];
	   g->ball_velocity[YY] = vy + g->paddle1_velocity[YY];
	 }

  /* Handle ball paddle2 collisions. */
  ball_paddle_distance = 0.0;
  for ( i = 0; i < 2; i++ )
	ball_paddle_distance +=
	  (g->ball[i] - g->paddle2[i])*(g->ball[i] - g->paddle2[i]);
  if ( ball_paddle_distance < g->bp_threshold )
	 {
	   g->status_bits |= PADDLE_TWO_CONTACT;
	   ball_paddle_distance = sqrt( (double) ball_paddle_distance );

	   /* Okay, rotate everything so collision is horizontal from right. */
	   
	   /* cosine angle */
	   c = (g->ball[XX] - g->paddle2[XX])/ball_paddle_distance; 
	   /* sin angle */
	   s = (g->ball[YY] - g->paddle2[YY])/ball_paddle_distance; 

	   vx = g->ball_velocity[XX] - g->paddle2_velocity[XX];
	   vy = g->ball_velocity[YY] - g->paddle2_velocity[YY];

	   v_rotated_x = c*vx + s*vy;
	   v_rotated_y = -s*vx + c*vy;

	   /* collision flips velocity */
	   if ( v_rotated_x < 0.0 )
	 v_rotated_x = -g->restitution*v_rotated_x;

	   vx = c*v_rotated_x - s*v_rotated_y;
	   vy = s*v_rotated_x + c*v_rotated_y;

	   g->ball_velocity[XX] = vx + g->paddle2_velocity[XX];
	   g->ball_velocity[YY] = vy + g->paddle2_velocity[YY];
	 }
  if ( game->status_bits & PADDLE_TWO_CONTACT ) {	
	  float tmp[2];
	  generate_paddle2_motion_imitate( game, tmp );
  }
}


/*****************************************************************************/
/*****************************************************************************/
void
generate_paddle2_motion_x( Game *g, float *paddle2_target )
{
  /* paddle2_target[XX] = 0.8*g->ball[XX]; */
  paddle2_target[YY] = g->top/2;
}
/*****************************************************************************/

void
generate_paddle2_motion_xy( Game *g, float *paddle2_target )
{
  float weighty = 0.03f;
  float weightx = 0.3f;
  float real_y_target;

  /* some defaults */
  paddle2_target[XX] = g->paddle2[XX];
  paddle2_target[YY] = g->paddle2[YY];

  if ( g->ball_velocity[YY] < 0.0 )
	real_y_target = /* g->paddle2_radius */ g->top/2;
  else
	real_y_target = g->paddle2_radius + 0.5*g->ball_velocity[YY];
  if ( real_y_target > g->top - g->paddle2_radius )
	real_y_target = g->top - g->paddle2_radius;
  if ( real_y_target < g->ball[YY] + g->ball_radius )
	real_y_target = g->ball[YY] + g->ball_radius;

  if ( real_y_target < g->ball[YY] + g->ball_radius )
	paddle2_target[YY] = weightx*real_y_target + (1-weightx)*g->paddle2[YY];
  else
	paddle2_target[YY] = weighty*real_y_target + (1-weighty)*g->paddle2[YY];

  if ( g->ball[YY] < g->paddle2[YY] )
	paddle2_target[XX] = g->paddle2[XX];
  else if ( g->ball[XX] < 0.0 )
	paddle2_target[XX] = weightx*g->right/2 + (1-weightx)*g->paddle2[XX];
  else
	paddle2_target[XX] = weightx*g->left/2 + (1-weightx)*g->paddle2[XX];
  
  if( sqrt( pow((board_length_div2-paddle2_target[YY]),2) +
		pow((paddle2_target[XX]),2) ) >g->max_reach) {
	paddle2_target[YY]=board_length_div2- 
	  sqrt( pow(g->max_reach,2)-pow((paddle2_target[XX]),2) );
  }
}

/*****************************************************************************/
/*****************************************************************************/

void
replay( Communication_Buffer *cb )
{
  if ( sample_ptr - samples > NPOINTS )
	sample_ptr = samples;

  cb->time = sample_ptr->time;
  cb->ball[0] = sample_ptr->ball[0];
  cb->ball[1] = sample_ptr->ball[1];
  cb->paddle1[0] = sample_ptr->paddle1[0];
  cb->paddle1[1] = sample_ptr->paddle1[1];
  cb->paddle2[0] = sample_ptr->paddle2[0];
  cb->paddle2[1] = sample_ptr->paddle2[1]; 
  cb->status_bits = sample_ptr->status_bits;

  sample_ptr++;
}

/*****************************************************************************/
/*****************************************************************************/

int last_data_file = 0;

void
write_samples()
{
  FILE *stream;
  char name[1000];

  stream = fopen( "last-data-file", "r" );
  if ( stream == NULL )
	{
	  fprintf( stderr, "Can't open last-data-file for read.\n" );
	  exit( -1 );
	}
  fscanf( stream, "%d", &last_data_file );
  fclose( stream );

  stream = fopen( "last-data-file", "w" );
  if ( stream == NULL )
	{
	  fprintf( stderr, "Can't open last-data-file for write.\n" );
	  exit( -1 );
	}
  fprintf( stream, "%d", ++last_data_file );
  fclose( stream );

  sprintf( name, "d%05d", last_data_file );
  stream = fopen( name, "w" );
  if ( stream == NULL )
	{
	  fprintf( stderr, "Can't open %s for write.\n", name );
	  exit( -1 );
	}
  printf( "Writing data to %s\n", name );
  for( sample_ptr = samples; sample_ptr - samples < NPOINTS; sample_ptr++ )
	{
	  fprintf( stream, "%d %g %g %g %g %g %g %g\n",
		   sample_ptr->status_bits,
		   sample_ptr->time,
		   sample_ptr->ball[0],
		   sample_ptr->ball[1],
		   sample_ptr->paddle1[0],
		   sample_ptr->paddle1[1],
		   sample_ptr->paddle2[0],
		   sample_ptr->paddle2[1]);
	  /*   fprintf( stream, "%lf %lf %lf %lf\n",
	   sample_ptr->ballV[0],
	   sample_ptr->ballV[1],
	   sample_ptr->paddle1V[0],
	   sample_ptr->paddle1V[1] 
	   );*/
	}
  fclose( stream );
  printf( "done.\n" );
}

/*****************************************************************************/

void
read_samples( char *filename )
{
  FILE *stream;

  stream = fopen( filename, "r" );
  if ( stream == NULL )
	{
	  fprintf( stderr, "Can't open %s for read.\n", filename );
	  exit( -1 );
	}
  printf( "Reading data from %s\n", filename );
  for( sample_ptr = samples; sample_ptr - samples < NPOINTS; sample_ptr++ )
	{
	  fscanf( stream, "%d %g %g %g %g %g %g %g\n",
		  &(sample_ptr->status_bits),
		  &(sample_ptr->time),
		  &(sample_ptr->ball[0]),
		  &(sample_ptr->ball[1]),
		  &(sample_ptr->paddle1[0]),
		  &(sample_ptr->paddle1[1]),
		  &(sample_ptr->paddle2[0]),
		  &(sample_ptr->paddle2[1]) );
	}
  fclose( stream );
  printf( "done.\n" );
}

/*****************************************************************************/
/*****************************************************************************/

/* Parameters that describe start of segment. */
float segment_x, segment_y, segment_time, segment_vx, segment_vy;
Hit *h;
Hit hh;

/* Parameter used in planning hit. */
float distance_so_far = 0.0;
float last_x;
float last_y;

void
generate_paddle2_motion_imitate( Game *g, float *paddle2_desireds )
{
  
  /* Have we hit the ball yet. */
  static int done_hit = 0;

  /* Have we planned this hit yet? */
  static int planned_this_hit_yet = FALSE;
  static int in_segment = FALSE; // Is the ball on the cyber humans side?
  /* some defaults */
  paddle2_desireds[XX] = g->paddle2[XX];
  paddle2_desireds[YY] = g->paddle2[YY];
  paddle2_desireds[XX+2] = 0.0;
  paddle2_desireds[YY+2] = 0.0;
  if ( g->ball[YY] < 0.0 ) {
	in_segment = FALSE; // The ball is on the human's side.
  }
  
  if ( g->status_bits & PADDLE_TWO_CONTACT ) {	
	//dcb_plan_hit( g, &hh, paddle2_desireds, 2 );//Follow through.
	g_cb->track_ball->Go(g->time); // Start the trackball moving.
	done_hit = TRUE; // Signal that the hit has been made.
	return;
  }
  if ( !in_segment ) {
	  //Set up for computing the paddle movement.
	//generate_paddle2_motion_simple( g, paddle2_desireds );
	  if ( g->ball[YY] > 0.0 ) {   // This is the first time it crossed the line.
		in_segment = TRUE;
		max_paddle_velocity=3.0;
	  /* Signs flipped on everything since these are player 2 variables. */
	  segment_x = -g->ball[XX];
	  segment_y = -g->ball[YY];
	  segment_vx = -g->ball_velocity[XX];
	  segment_vy = -g->ball_velocity[YY];

	  // Get the type of hit that is to be made.
	  h = lookup( segment_x, segment_y, segment_vx, segment_vy );
	  if ( h->type == 0 ) {
		printf( "NO HIT.\n" );
	  } else {
		/* Flipping signs to put hit in paddle2's point of view */
		hh.type = 1;
		hh.hit_x = -h->hit_x;
		hh.hit_y = -h->hit_y;
		hh.hit_t = h->hit_t;
		hh.hit_d = h->hit_d;
		hh.hit_v = h->hit_v;
		hh.target_x = -h->target_x;
		hh.target_y = -h->target_y;
		distance_so_far = 0.0;
		last_x = g->ball[XX];
		last_y = g->ball[YY];
		done_hit = 0;
		planned_this_hit_yet = FALSE;
	  }
	} else {  // The paddle is on the human's side.
		static count=0;
		idle_point * idle_ptr;
		idle_ptr=idle.lookupDWAM(-g->ball[XX], -g->ball[YY],  -g->ball_velocity[XX], -g->ball_velocity[YY]);
		paddle2_desireds[YY]=-idle_ptr->Py; //Place paddle.
		paddle2_desireds[XX]=-idle_ptr->Px; //Place paddle.
		if(0 && count++==30) {
			count=0;
			cout << "Xi:"<<-g->ball[XX]<<" Yi:"<<-g->ball[YY]<< "    X:"<<idle_ptr->Px<<" Y:"<<idle_ptr->Py<<endl;
		}
		max_paddle_velocity=2.0;  // Slow the paddle down so it looks more natural.

	  //paddle2_desireds[YY]=0.35f; //Place paddle.
	  //paddle2_desireds[XX]=0.0; //Place paddle.
	  //dcb_idle( g, &hh, paddle2_desireds, 1 );
	  return;
	}
  } else if ( h->type == 1 && !done_hit ) { // The ball is on the cyber player's side.
	  // The ball was hit by the human in this situation and has not yet been hit by the cyber player.
	if ( planned_this_hit_yet == FALSE ) {
	  dcb_procedure_hit( g, &hh, paddle2_desireds, 1 );
	  //dcb_plan_hit( g, &hh, paddle2_desireds, 1 );
	  //plan_hit( g, &hh, paddle2_desireds, 1 );
	  planned_this_hit_yet = TRUE;  // signal that a hit was planned.
	} else {
	  dcb_procedure_hit( g, &hh, paddle2_desireds, 0 );
	//dcb_plan_hit( g, &hh, paddle2_desireds, 0 );
	//plan_hit( g, &hh, paddle2_desireds, 0 );
	}
  }
}

/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/
/*
 *
 * Plan the position of the paddle to make the hit described by h.
 *
 */
void
dcb_plan_hit( Game *g, Hit *h, float *paddle2_desireds, int first_time )
{
  int i;
  //static int planned;
  static float last_time;
  float increment=0.05f; //The time interval used for predicting the future.
  float ball[2];
  float i_ball[2],i_ball_velocity[2];
  float n_ball[2],n_ball_velocity[2];
  static float ball_x, ball_y;
  static float target_x, target_y;
  float ball_v=0, error, distance;
  static float des_x, des_y, time_to_impact;
  float hit_dist=h->hit_d; //The distance the ball will travel from the center
  //line to the point it is hit.

  static float p_hit_vel; //The desired vel. of the paddle when the ball is hit.
  float hit_vel=h->hit_v; //The desired vel. of the ball after it is hit.
  float shot_time=0.4f; //The time spent while actually hitting the ball.
  //float setup_dist; //The distance from the impact point to obtain the desired
  //velocity in the desired amount of time.

  // hit_vel=2;
  /* Get current ball state */
  for( i = 0; i < 2; i++ ) {
	ball[i] = g->ball[i];
	i_ball[i] = g->ball[i];
	i_ball_velocity[i] = g->ball_velocity[i];
  }

  if(first_time==1) {
	distance_so_far=0;
	last_x = ball[XX];
	last_y = ball[YY];
	last_time=g->time;
	time_to_impact=9999;
	//planned = 0; //The hit has not yet been planned.
	return;
  }

  /* Maintain distance_so_far variable. */
  //Does not take wall hits into account.
  distance_so_far += sqrt( (double)
			   (ball[XX] - last_x)*
			   (ball[XX] - last_x) +
			   (ball[YY] - last_y)*
			   (ball[YY] - last_y) );

  error=hit_dist-distance_so_far;

  if(first_time==2) {
	/////////////////////
	//Here is the cheat.
	g->ball_velocity[XX]=-hit_vel*cos(theta_des);
	g->ball_velocity[YY]=-hit_vel*sin(theta_des);
	return;
  }

  //Find the time to the hit point by projecting the ball into the future.
  distance=0;
  if(error >0)
	time_to_impact=0;
  if(error < -0.02) { //The ball was missed.
	return;
  }
  while( error>0) { 
	if(error<0.2) increment=0.01f;
	if(error<0.1) increment=0.001f;
	//ends on the last time before actual impact.
	time_to_impact+=increment;
	
	do_dynamics_collisions(increment, i_ball, i_ball_velocity,
			   n_ball, n_ball_velocity);
	
	distance+=sqrt(square(n_ball[XX]-i_ball[XX])+
		   square(n_ball[YY]-i_ball[YY]));
	if(error == hit_dist-distance-distance_so_far) {
	  //The ball has stopped moving
	  return;
	}
	error=hit_dist-distance-distance_so_far;
	
	//printf("++ X:%0.3f X1:%0.3f  V:%0.3f V1:%0.3f\n", 
	//	 i_ball[XX], n_ball[XX], i_ball_velocity[XX], n_ball_velocity[XX]);
	i_ball[XX]=n_ball[XX];
	i_ball[YY]=n_ball[YY];
	i_ball_velocity[XX] = n_ball_velocity[XX];
	i_ball_velocity[YY] = n_ball_velocity[YY];
  }

  //The new ball velocity with friction at the time of impact.
  ball_v=sqrt(square(i_ball_velocity[XX])+square(i_ball_velocity[YY]));

  if(i_ball_velocity[YY] <0) {	
	//The ball hit the back wall within the hit distance.
	return;
  }
  
  //Get travel angle relative to center line.
  theta=atan2(i_ball_velocity[YY], i_ball_velocity[YY]);
  
  //position of ball at hit_dist from center line.	  
  ball_y=i_ball[YY];
  ball_x=i_ball[XX];

  //Get angle from hit point to center of the goal.
  theta_des=atan2(board_length_div2+ball_y, ball_x);

  //Ensure the output velocity is not less than the input velocity
  if(hit_vel<ball_v) {
	hit_vel=ball_v;
  }

  //////////////////////////////
	
  //Get travel angle relative to center line.
  theta=atan2(i_ball_velocity[YY], i_ball_velocity[XX]);
  
  if(h->target_x < -0.2) {
	//right bank.
	target_x=ball_x+BOARD_WIDTH+0.04;
	target_y=board_length_div2+ball_y;
  } else if(h->target_x > 0.2) {
	//left bank.
	target_x=ball_x-BOARD_WIDTH-0.04;
	target_y=board_length_div2+ball_y;
  } else {
	//straight in shot.
	target_x=ball_x;
	target_y=board_length_div2+ball_y;
  }

  //Get angle from hit point to thetarget.
  theta_des=atan2(target_y, target_x);

  //Compute OX and OY using theta_des and desired velocity.
  OVx=-hit_vel*cos(theta_des);
  OVy=-hit_vel*sin(theta_des);

  IVx=i_ball_velocity[XX];
  IVy=i_ball_velocity[YY];

  float a, b;
  //Compute  a and b.
  a=(1+g->restitution)*((square(OVx-IVx))+ (square((OVy-IVy))));

  b=OVx*(square(OVx-IVx)) + (OVx-IVx)*OVy*(OVy-IVy) + 
	IVx*g->restitution*(square(OVx-IVx)) + 
	IVy*g->restitution*(OVx-IVx)*(OVy-IVy);
	
  //compute PY and PX.
  if(IVx==0 && OVx==0) {
	PVx=0;
	PVy=(square(OVy+IVy))/(1+g->restitution);
  } else {
	PVx=b/a;
	PVy=PVx*(OVy-IVy)/(OVx-IVx);
  }
  p_hit_vel=sqrt(square(PVx)+square(PVy));

  //Get the hit angle.
  if(p_hit_vel>0.05 && PVy<0)//Do not try to hit the ball backwards.
	theta_hit=atan2(PVy, PVx); 
  else
	theta_hit = (theta_des+theta)/2; //For stationary paddle
  
  /////////////////////////
  if(distance == 0) {
	//	 printf("******************\n");
	time_to_impact -= (g->time-last_time);
  }
  float shot_comp=0.033f; //Puts hit one cycle ahead in time.
  //float shot_comp=0.0;
  if(time_to_impact<shot_time) { //Make the shot.
	des_x=ball_x-(g->paddle2_radius+g->ball_radius+
		  (time_to_impact-shot_comp)*p_hit_vel)*cos(theta_hit);
	des_y=ball_y-(g->paddle2_radius+g->ball_radius+
		  (time_to_impact-shot_comp)*p_hit_vel)*sin(theta_hit);
  } else { //Just place the paddle in preparation for the shot.
	des_x=ball_x-(g->paddle2_radius+g->ball_radius+
		  (shot_time-shot_comp)*p_hit_vel)*cos(theta_hit);
	des_y=ball_y-(g->paddle2_radius+g->ball_radius+
		  (shot_time-shot_comp)*p_hit_vel)*sin(theta_hit);
  }


  //printf("\nbx:%.3f, by:%.3f, dx:%.3f dy:%.3f\n", ball_x, ball_y,
  // des_x, des_y);

  //getchar();

  //des_x=ball_x;
  //des_y=ball_y;
  //if(des_y > board_length_div2-PADDLE_RADIUS) {
  //des_y = board_length_div2-PADDLE_RADIUS;
  //}
  if ( des_x < g->left + g->paddle2_radius ) {
	des_x=g->left + g->paddle2_radius;
  }
  if ( des_x > g->right - g->paddle2_radius ) {
	des_x=g->right + g->paddle2_radius;
  }
  if ( des_y < g->bottom + g->paddle2_radius ) {
	des_y= g->bottom + g->paddle2_radius;
  }
  if ( des_y >	board_length_div2- g->paddle2_radius ) {
	des_y= board_length_div2 - g->paddle2_radius;
  }


  //printf("TH:%0.3lf Len:%0.3lf, cx:%0.3lf\n", theta_hit, length, center_x );


  if( sqrt( pow((board_length_div2-des_y),2) +
		pow((des_x),2) ) >g->max_reach) {
	des_y=board_length_div2- 
	  sqrt( pow(g->max_reach,2)-pow((des_x),2) );
	//printf("newval: %.3f reach: %.3f\n", des_y, g->max_reach);
  }

  //printf("++ Pad vel PVx:%0.6f PVy:%0.3f  act x:%0.3f y:%0.3f, bvel:%.3f\n", 
  // PVx, PVy, g->paddle2_velocity[XX], g->paddle2_velocity[YY],
  //	 des_y-g->paddle2[YY]);

  //  printf("++ Ball vel t:%0.6f  act x:%0.3f y:%0.3f\n", 
  // g->time, OVx, OVy);

  // printf("bx:%.3f, by:%.3f, dx:%.3f dy:%.3f\n", ball_x, ball_y,
  // des_x, des_y);
  paddle2_desireds[0]=des_x;
  paddle2_desireds[1]=des_y;
  
  //printf("x: %.3f  y: %.3f\n", PX, PY);

  last_x = ball[XX];
  last_y = ball[YY];
  last_time=g->time;
}

/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/
/*
 *
 * Plan the position of the paddle to make the hit described by h.
 *
 */
void
dcb_procedure_hit( Game *g, Hit *h, float *paddle2_desireds, int first_time )
{
  int i;
  static int planned;
  static float last_time;
  float increment=0.05f; //The time interval used for predicting the future.
  float ball[2];
  float i_ball[2],i_ball_velocity[2];
  float n_ball[2],n_ball_velocity[2];
  static float ball_x, ball_y, ball_vx, ball_vy;
  static float target_x, target_y;
  float  error, distance, theta_des;
  static float des_x, des_y, time_to_impact;
  float hit_dist=h->hit_d; //The distance the ball will travel from the center
  //line to the point it is hit.

  float hit_vel=h->hit_v; //The desired vel. of the ball after it is hit.
  float shot_time=0.4f; //The time spent while actually hitting the ball.
  //hit_vel=2;

  /* Get current ball state */
  for( i = 0; i < 2; i++ ) {
	ball[i] = g->ball[i];
	i_ball[i] = g->ball[i];
	i_ball_velocity[i] = g->ball_velocity[i];
  }

  if(first_time==1) {
	distance_so_far=0;
	last_x = ball[XX];
	last_y = ball[YY];
	last_time=g->time;
	time_to_impact=9999;
	planned = 0; //The hit has not yet been planned.
	g_cb->track_ball->Reset();

	target_y=-board_length_div2+BALL_RADIUS;
	if(h->target_x < -0.2) {
	  //right bank.
	  //cout << "** Right Bank Shot\07" << endl;
	  cout << "** Right Bank Shot" << endl;
	  target_x=-BOARD_WIDTH;
	} else if(h->target_x > 0.2) {
	  //left bank.
	  cout << "** Left Bank Shot" << endl;
	  target_x=BOARD_WIDTH;
	} else {
	  //straight in shot.
	  cout << "** Straight Shot" << endl;
	  target_x=0;
	}
	return;
  }

  /* Maintain distance_so_far variable. */
  //Does not take wall hit errors into account.
  distance_so_far += sqrt( (double)
			   (ball[XX] - last_x)*
			   (ball[XX] - last_x) +
			   (ball[YY] - last_y)*
			   (ball[YY] - last_y) );

  planned--;
  if(planned < 0) {
	planned =4; //Plan once every x time.

	error=hit_dist-distance_so_far;

	//Find the time to the hit point by projecting the ball into the future.
	distance=0;
	if(error >0)
	  time_to_impact=0;
	if(error < -0.02) { //The ball was missed.
	  return;
	}
	while( error>0) { 
	  if(error<0.2) increment=0.01f;
	  if(error<0.02) increment=0.001f;  // Use smaller steps near the paddle.
	  //ends on the last time before actual impact.
	  time_to_impact+=increment;
	  
	  do_dynamics_collisions(increment, i_ball, i_ball_velocity,
				 n_ball, n_ball_velocity);
	
	  distance+=sqrt(square(n_ball[XX]-i_ball[XX])+
			 square(n_ball[YY]-i_ball[YY]));
	  if(fabs(error - (hit_dist-distance-distance_so_far))<0.00001) {
		//The ball has stopped moving
		return;
	  }
	  error=hit_dist-distance-distance_so_far;
	
	  i_ball[XX]=n_ball[XX];
	  i_ball[YY]=n_ball[YY];
	  i_ball_velocity[XX] = n_ball_velocity[XX];
	  i_ball_velocity[YY] = n_ball_velocity[YY];
	}
	//position of ball at hit_dist from center line.	
	ball_y=i_ball[YY];
	ball_x=i_ball[XX];
	//velocity of ball at hit_dist from center line.	
	ball_vy=i_ball_velocity[YY];
	ball_vx=i_ball_velocity[XX];

	Procedure  * pr_ptr;
	float pb_dist = g->ball_radius + g->paddle1_radius;

	if (use_nn) {
	  pr_ptr = procedures.lookupNN(-ball_x, -ball_y,-ball_vx,
				   -ball_vy, -target_x, -target_y, hit_vel);
	} else {
	  pr_ptr = procedures.lookupDWAM(-ball_x, -ball_y,-ball_vx,
					 -ball_vy, -target_x, -target_y, hit_vel);
	}
	PBx=-pb_dist*sin(pr_ptr->PB_theta);
	PBy=-pb_dist*cos(pr_ptr->PB_theta);
	PVx=-pr_ptr->PV_x;
	PVy=-pr_ptr->PV_y;
  } else {
	time_to_impact-=g->time-last_time;
  }

  //Get angle from hit point to thetarget.
  theta_des=atan2(target_y-ball_y, target_x-ball_x);
  if (ball_vy>0){
	g_cb->track_ball->Initialize(time_to_impact+g->time,
				   ball_x, ball_y, hit_vel*cos(theta_des),
				   hit_vel*sin(theta_des));
  }

  if(ball_vy <0) {	
	//The ball hit the back wall within the hit distance.
	return;
  }
  des_x=ball_x+PBx;
  des_y=ball_y+PBy;

  float shot_comp=0.033f;
  //float shot_comp=0.015; //Approx half a time cycle.
  if(time_to_impact<shot_time) { //Make the shot.
	des_x=ball_x-(PBx + (time_to_impact-shot_comp)*PVx);
	des_y=ball_y-(PBy + (time_to_impact-shot_comp)*PVy);
  } else { //Just place the paddle in preparation for the shot.
	des_x=ball_x-(PBx + (shot_time-shot_comp)*PVx);
	des_y=ball_y-(PBy + (shot_time-shot_comp)*PVy);
  }

  if(des_y > board_length_div2-PADDLE_RADIUS) {
	des_y = board_length_div2-PADDLE_RADIUS;
  }
  if ( des_x < g->left + g->paddle2_radius ) {
	des_x=g->left + g->paddle2_radius;
  }
  if ( des_x > g->right - g->paddle2_radius ) {
	des_x=g->right + g->paddle2_radius;
  }
  if ( des_y < g->bottom + g->paddle2_radius ) {
	des_y= g->bottom + g->paddle2_radius;
  }
  if ( des_y >	board_length_div2- g->paddle2_radius ) {
	des_y= board_length_div2 - g->paddle2_radius;
  }

  if( sqrt( pow((board_length_div2-des_y),2) +
		pow((des_x),2) ) >g->max_reach) {
	des_y=board_length_div2- 
	  sqrt( pow(g->max_reach,2)-pow((des_x),2) );
  }

  paddle2_desireds[0]=des_x;
  paddle2_desireds[1]=des_y;
  
  last_x = ball[XX];
  last_y = ball[YY];
  last_time=g->time;
}
/*****************************************************************************/
/*****************************************************************************/

#define PLANNING_SIMULATOR_TIME_INCREMENT 0.01f

float last_paddle2_desireds[4];
float last_planning_time;

void
plan_hit( Game *g, Hit *h, float *paddle2_desireds, int first_time )
{
  float target[2];
  float target_vector[2];
  float bisector[2];

  /* Planning state */
  float time;
  float last_time;
  float ball[2];
  float ball_previous_location[2];
  float ball_velocity[2];
  float new_v;
  float time_increment = PLANNING_SIMULATOR_TIME_INCREMENT;
  int i;
  float distance = 0.0;
  float s1, s2, s3, s4;
  int target_adjusted = 0;
  float target_x;
  float incoming_velocity[2];
  float outgoing_velocity[2];
  float paddle_velocity[2];
  float velocity;
  float planning_time_increment;
  float time_to_impact;
  float position_error[2];

  /* Figure out timing interval */
  if ( first_time ) {
	planning_time_increment = 0.03f;
  }
  else {
	planning_time_increment = g->time - last_planning_time;
  }
  last_planning_time = g->time;

  /* Get current ball state */
  for( i = 0; i < 2; i++ ) {
	ball[i] = g->ball[i];
	ball_velocity[i] = g->ball_velocity[i];
  }

  /* Maintain distance_so_far variable. */
  distance_so_far += sqrt( (double)
			  (ball[XX] - last_x)*
			  (ball[XX] - last_x) +
			  (ball[YY] - last_y)*
			  (ball[YY] - last_y) );
  last_x = ball[XX];
  last_y = ball[YY];
  distance = distance_so_far;

  /* Go to high resolution mode if close to impact. */
  velocity = sqrt( (double) ball_velocity[XX]*ball_velocity[XX] +
		  ball_velocity[YY]*ball_velocity[YY] );
  time_to_impact = (h->hit_d - distance)/velocity;
  if ( time_to_impact < PLANNING_SIMULATOR_TIME_INCREMENT )
	time_increment = (float) (PLANNING_SIMULATOR_TIME_INCREMENT/10.0);

  /* Integrate forward to find hit location. */
  last_time = 0.0;
  for( time = time_increment; distance < h->hit_d; time += time_increment ) {
	for( i = 0; i < 2; i++ ) {
	  ball_previous_location[i] = ball[i];
	  new_v = ball_velocity[i]*(1.0 - g->friction*(time - last_time));
	  ball[i] += 0.5*(ball_velocity[i] + new_v)*(time - last_time);
	  ball_velocity[i] = new_v;
	}
	last_time = time;

	/* Handle ball wall collisions. */
	if ( ball[XX] < g->left + g->ball_radius ) //Left side hit.
	  {
	ball[XX] = 2*(g->left + g->ball_radius) - ball[XX];
	ball_velocity[XX] = -g->restitution*ball_velocity[XX];
	  }
	if ( ball[XX] > g->right - g->ball_radius ) {
	  ball[XX] = 2*(g->right - g->ball_radius) - ball[XX];
	  ball_velocity[XX] = -g->restitution*ball_velocity[XX];
	}
	if ( ball[YY] < g->bottom + g->ball_radius ) {
	  ball[YY] = 2*(g->bottom + g->ball_radius) - ball[YY];
	  ball_velocity[YY] = -g->restitution*ball_velocity[YY];
	}
	if ( ball[YY] > g->top - g->ball_radius ) {
	  ball[YY] = 2*(g->top - g->ball_radius) - ball[YY];
	  ball_velocity[YY] = -g->restitution*ball_velocity[YY];
	}

	distance += sqrt( (double)
			  (ball[XX] - ball_previous_location[XX])*
			  (ball[XX] - ball_previous_location[XX]) +
			  (ball[YY] - ball_previous_location[YY])*
			  (ball[YY] - ball_previous_location[YY]) );

	/* Go to high resolution mode if close to impact. */
	velocity = sqrt( (double) ball_velocity[XX]*ball_velocity[XX] +
			 ball_velocity[YY]*ball_velocity[YY] );
	time_to_impact = (h->hit_d - distance)/velocity;
	if ( time_to_impact < PLANNING_SIMULATOR_TIME_INCREMENT )
	  time_increment = (float) (PLANNING_SIMULATOR_TIME_INCREMENT/10.0);
  }
  target[XX] = 0.0;
  target[XX] = -0.5;
  target[XX] = 0.5;
  target[YY] = g->bottom;

  /* set target */
  target_x = h->target_x;
  target_adjusted = 0;
  for( ; ; ) {
	i = target_adjusted;
	if ( target_x > board_width_div2 ) {
	  target[XX] += BOARD_WIDTH;
	  target_x -= BOARD_WIDTH;
	  target_adjusted++;
	}
	if ( target_x < -board_width_div2 ) {
	  target[XX] -= BOARD_WIDTH;
	  target_x += BOARD_WIDTH;
	  target_adjusted++;
	}
	if ( target_adjusted == i )
	  break;
  }
  /* Okay, intercept point and targets are selected. **********************/
  
  /* Hit with specified velocity version **********************************/

  /* first figure out paddle velocity. */
  /* Incoming velocity is just ball_velocity */
  for( i = 0; i < 2; i++ )
	incoming_velocity[i] = ball_velocity[i];

  /* What is outgoing velocity? */
  for( i = 0; i < 2; i++ )
	target_vector[i] = target[i] - ball[i];
  s1 = normalize_vector( target_vector, 2 );

  for( i = 0; i < 2; i++ )
	outgoing_velocity[i] = h->hit_v*target_vector[i];
  
  for( i = 0; i < 2; i++ )
	paddle_velocity[i] = (incoming_velocity[i] + outgoing_velocity[i])/2;

  /* Now we need to figure out the bisector of the two velocities
	 after the paddle velocity is subtracted out. */
  for( i = 0; i < 2; i++ ) {
	incoming_velocity[i] -= paddle_velocity[i];
	outgoing_velocity[i] -= paddle_velocity[i];
  }
  s2 = normalize_vector( incoming_velocity, 2 );
  s3 = normalize_vector( outgoing_velocity, 2 );
  for( i = 0; i < 2; i++ )
	bisector[i] = outgoing_velocity[i] - incoming_velocity[i];
  s4 = normalize_vector( bisector, 2 );
  
  if ( s1 > 0.001 && s2 > 0.001 && s3 > 0.001 && s4 > 0.001 ) {
	if ( time > planning_time_increment ) {
	  for( i = 0; i < 2; i++ ) {
	position_error[i] =
	  ball[i] - (g->paddle2_radius + g->ball_radius)*bisector[i]
	  - time*paddle_velocity[i]- g->paddle2[i];
	paddle2_desireds[i] = g->paddle2[i];
	paddle2_desireds[i+2] = paddle_velocity[i]
	  + position_error[i]/planning_time_increment;
	  }
	} else {
	  for( i = 0; i < 2; i++ ){
	paddle2_desireds[i] = g->paddle2[i];
	paddle2_desireds[i+2] = paddle_velocity[i];
	position_error[i] = 0;
	  }
	}
  } else {
	for( i = 0; i < 2; i++ )
	  paddle2_desireds[i] = last_paddle2_desireds[i];
	printf( "%g: %g %g vectors too small\n", time,
		paddle2_desireds[0], paddle2_desireds[1] );
  }

  /* END Hit with specified velocity version ******************************/
/* #endif */

  debug_array[debug_index][0] = position_error[0];
  debug_array[debug_index][1] = position_error[1];

  debug_array[debug_index][2] = paddle_velocity[0];
  debug_array[debug_index][3] = paddle_velocity[1];

  debug_array[debug_index][4] = time;
  debug_array[debug_index][5] = ball[0];
  debug_array[debug_index][6] = ball[1];
  debug_array[debug_index][7] =
	ball[0] - (g->paddle2_radius + g->ball_radius)*bisector[0];
  debug_array[debug_index][8] = 
	ball[1] - (g->paddle2_radius + g->ball_radius)*bisector[1];
  for( i = 0; i < 2; i++ )
	outgoing_velocity[i] = h->hit_v*target_vector[i];

  for( i = 0; i < 4; i++ )
	last_paddle2_desireds[i] = paddle2_desireds[i];
}


/*****************************************************************************/

void
generate_paddle2_motion_target( Game *g, float *paddle2_target )
{
  float intercept_line;
  float dt;
  float x_predicted;
  float target[2];
  float target_vector[2];
  float ball_vector[2];
  float bisector[2];

  target[XX] = 0.0;
  target[YY] = g->bottom;

  intercept_line = g->top/2;

  if ( g->ball_velocity[YY] < 0.0 || g->ball[YY] < 0.0 )
	{
	  paddle2_target[XX] = g->paddle2[XX];
	  return;
	}

  /* Predict where ball is going to intercept half line. */
  dt = (intercept_line - g->ball[YY])/g->ball_velocity[YY];

  x_predicted = g->ball[XX] + g->ball_velocity[XX]*dt;
  if ( x_predicted < g->left || x_predicted > g->right )
	{
	  paddle2_target[XX] = g->paddle2[XX];
	  return;
	}

  target_vector[XX] = target[XX] - x_predicted;
  target_vector[YY] = target[YY] - intercept_line;
  normalize_vector( target_vector, 2 );

  ball_vector[XX] = x_predicted - g->ball[XX];
  ball_vector[YY] = intercept_line - g->ball[YY];
  normalize_vector( ball_vector, 2 );

  bisector[XX] = target_vector[XX] - ball_vector[XX];
  bisector[YY] = target_vector[YY] - ball_vector[YY];
  normalize_vector( bisector, 2 );

  paddle2_target[XX] = x_predicted - g->paddle2_radius*bisector[XX];
  paddle2_target[YY] = intercept_line - g->paddle2_radius*bisector[YY];
}

/*****************************************************************************/

float
normalize_vector( float *vector, int length )
{
  int i;
  double sum = 0.0;
  float scale_factor;

  for( i = 0; i < length; i++ )
	{
	  sum += vector[i]*vector[i];
	}
  sum = sqrt( sum );
  scale_factor = 1.0/sum;
  for( i = 0; i < length; i++ )
	{
	  vector[i] *= scale_factor;
	}
  return sum;
}

/*****************************************************************************/

void
generate_paddle2_motion_simple( Game *g, float *paddle2_desireds )
{
  float weighty = 0.03f;
  float weightx = 0.3f;
  float real_y_target;

  /* some defaults */
  paddle2_desireds[XX] = g->paddle2[XX];
  paddle2_desireds[YY] = g->paddle2[YY];
  paddle2_desireds[XX+2] = 0.0;
  paddle2_desireds[YY+2] = 0.0;

  if ( g->ball_velocity[YY] < 0.0 )
	real_y_target = /* g->paddle2_radius */ g->top/2;
  else
	real_y_target = g->paddle2_radius + 0.5*g->ball_velocity[YY];
  if ( real_y_target > g->top - g->paddle2_radius )
	real_y_target = g->top - g->paddle2_radius;
  if ( real_y_target < g->ball[YY] + g->ball_radius )
	real_y_target = g->ball[YY] + g->ball_radius;

  if ( real_y_target < g->ball[YY] + g->ball_radius )
	paddle2_desireds[YY] = weightx*real_y_target + (1-weightx)*g->paddle2[YY];
  else
	paddle2_desireds[YY] = weighty*real_y_target + (1-weighty)*g->paddle2[YY];

  if ( g->ball[YY] < g->paddle2[YY] )
	paddle2_desireds[XX] = g->paddle2[XX];
  else if ( g->ball[XX] < 0.0 )
	paddle2_desireds[XX] = weightx*g->right/2 + (1-weightx)*g->paddle2[XX];
  else
	paddle2_desireds[XX] = weightx*g->left/2 + (1-weightx)*g->paddle2[XX];

  /*
  paddle2_desireds[XX] = g->paddle2[XX];
  */
}

/*****************************************************************************/
/*****************************************************************************/

#define MAX_N_HITS 1000

Hit hits_database[MAX_N_HITS];
int hits_index = 0;

float hit_x_min = 1e10;
float hit_x_max = -1e10;
float hit_y_min = 1e10;
float hit_y_max = -1e10;
float hit_vx_min = 1e10;
float hit_vx_max = -1e10;
float hit_vy_min = 1e10;
float hit_vy_max = -1e10;
float x_scale, vx_scale, vy_scale;

/*****************************************************************************/
void
read_hits( char *filename )
{
  FILE *stream;
  float x, y, vx, vy;
  int type;

  stream = fopen( filename, "r" );
  if ( stream == NULL )
	{
	  printf( "No hits file: %s\n", filename );
	  return;
	}

  hits_index=0;
  for( ; ; )
	{
	  if ( hits_index >= MAX_N_HITS )
	break;
	  if ( fscanf( stream, "%g%g%g%g%d", &x, &y, &vx, &vy, &type ) < 5 )
	break;
	  hits_database[hits_index].x = x;
	  hits_database[hits_index].y = y;
	  hits_database[hits_index].vx = vx;
	  hits_database[hits_index].vy = vy;
	  hits_database[hits_index].type = type;
	  if ( x < hit_x_min )
	hit_x_min = x;
	  if ( x > hit_x_max )
	hit_x_max = x;
	  if ( y < hit_y_min )
	hit_y_min = y;
	  if ( y > hit_y_max )
	hit_y_max = y;
	  if ( vx < hit_vx_min )
	hit_vx_min = vx;
	  if ( vx > hit_vx_max )
	hit_vx_max = vx;
	  if ( vy < hit_vy_min )
	hit_vy_min = vy;
	  if ( vy > hit_vy_max )
	hit_vy_max = vy;
	  
	  if ( type == 1 )
	{
	  fscanf( stream, "%g%g%g%g%g%g%g",
		 &(hits_database[hits_index].hit_x),
		 &(hits_database[hits_index].hit_y),
		 &(hits_database[hits_index].hit_t),
		 &(hits_database[hits_index].hit_d),
		 &(hits_database[hits_index].hit_v),
		 &(hits_database[hits_index].target_x),
		 &(hits_database[hits_index].target_y) );
	}
	  else
	{
	  /* No additional parameters for case 0 */
	}
	  hits_index++;
	}

  fclose( stream );
  //printf( "%d hits read.\n", hits_index );
  cout << hits_index << " hits read.\n";
  x_scale = (hit_x_max - hit_x_min)*(hit_x_max - hit_x_min);
  vx_scale = (hit_vx_max - hit_vx_min)*(hit_vx_max - hit_vx_min);
  vy_scale = (hit_vy_max - hit_vy_min)*(hit_vy_max - hit_vy_min);
}

/*****************************************************************************/
/*****************************************************************************/

Hit *
lookup( float x, float y, float vx, float vy )
{
  int i;
  float distance;
  float best_distance = 1e10;
  int best_hit = -1;

  for( i = 0; i < hits_index; i++ ) {
	distance = 0.0;
	distance += (hits_database[i].x - x)*(hits_database[i].x - x)/
	  x_scale;
	/* ignore y */
	distance += (hits_database[i].vx - vx)*(hits_database[i].vx - vx)/
	  vx_scale;
	distance += (hits_database[i].vy - vy)*(hits_database[i].vy - vy)/
	  vy_scale;

	if ( distance < best_distance ) {
	  best_distance = distance;
	  best_hit = i;
	}
  }
  if ( best_hit == -1 )
	{
	  fprintf( stderr, "Screwup in lookup: %g %g %g %g\n", x, y, vx, vy );
	  exit( -1 );
	}
  //printf( "H %d Dist:%f\n", best_hit,hits_database[best_hit].hit_d);
  return &(hits_database[best_hit]);
}

/*****************************************************************************/
void
print_debug_array()
{
  int i, j;

  for( i = 0; i < debug_index; i++ )
	{
	  for( j = 0; j < N_DEBUG_DIMENSIONS; j++ )
	printf( "%5.3f ", debug_array[i][j] );
	  printf( "\n" );
	}
}

/*****************************************************************************/

