#include "track.hpp"
#include "air-hockey.h"
#include <iostream.h>

Track :: Track(SoSeparator * inRoot)
{
  //window = in_window;
  root = inRoot;
  ballRoot = new SoSeparator;
  ballMaterial = new SoMaterial;
  show = 0;
  z=BALL_HEIGHT/2;
  initialized=0;
  running=FALSE;

  SoSeparator   *ballCache       = new SoSeparator;
  SoCylinder     *ball            = new SoCylinder;

  ballTransform = new SoTransform;
  drawStyle = new SoDrawStyle;
  drawStyle->style=SoDrawStyle::INVISIBLE;

  ball->radius = BALL_RADIUS;
  ball->height = BALL_HEIGHT;
  ballMaterial->shininess = .8f;
  ballMaterial->ambientColor.setValue(SbColor(0.4f, 0.4f, 0.4f));
  ballMaterial->diffuseColor.setValue(SbColor(0.3f, 0.3f, 0.3f));
  ballMaterial->specularColor.setValue(SbColor(0.6f, 0.6f, 0.6f));
  ballCache->renderCaching = SoSeparator::ON;
  ballCache->addChild(ballMaterial);
  ballCache->addChild(ball);
  ballRoot->addChild(drawStyle);

  ballRoot->addChild(ballTransform);
  ballRoot->addChild(ballCache);
  root->addChild(ballRoot);

  ballTransform->
    rotation.setValue(SbVec3f(1.0, 0.0, 0.0), (float) (M_PI/2.0) );
  ballTransform->translation.setValue(0, 0, z);
}


void Track :: Toggle() {
  if(show) {
    drawStyle->style=SoDrawStyle::INVISIBLE;
    show=0;
  } else {
    drawStyle->style=SoDrawStyle::FILLED;
    show=1;
  }
};


void Track ::  Reset()
{
  drawStyle->style=SoDrawStyle::INVISIBLE;
  running=FALSE;
  time=0;
  initialized=0;
}

void Track ::  Go(float itime)
{
  if (running) //The ball is already moving.
    return;

  ballTransform->translation.setValue(x, y, z);
  if(show) {
    drawStyle->style=SoDrawStyle::FILLED;
  }
  running=TRUE;
  time=itime;
}

void Track ::  Initialize(float itime,float ix, float iy, 
		     float iVx, float iVy)
{
  if (running) //The ball is moving.
    return;
  time=itime;
  x=ix;
  y=iy;
  Vx=iVx*1.2;
  Vy=iVy;
  if(show) {
    drawStyle->style=SoDrawStyle::FILLED;
    initialized =TRUE;
  }
  ballTransform->translation.setValue(x, y, z);
}

void Track :: Simulate(float newTime)
{
  int i;
  float increment = (newTime-time)/10;
  float n_ball_velocity[3], i_ball_velocity[3];
  float i_ball[3], n_ball[3];

  if(!initialized || newTime<start_time || !running || !show)
    return;

  i_ball[XX]=x;
  i_ball[YY]=y;
  i_ball_velocity[XX]=Vx;
  i_ball_velocity[YY]=Vy;

  //Break segment up into ten steps.
  for(i=0;i<10;i++) {
    
    do_dynamics_collisions(increment, i_ball, i_ball_velocity,
			   n_ball, n_ball_velocity);
    i_ball[XX]=n_ball[XX];
    i_ball[YY]=n_ball[YY];
    i_ball_velocity[XX] = n_ball_velocity[XX];
    i_ball_velocity[YY] = n_ball_velocity[YY];
  }
  time = newTime;
  x  = i_ball[XX];
  y  = i_ball[YY];
  Vx = i_ball_velocity[XX];
  Vy = i_ball_velocity[YY];

  if(Vy>0) { //the ball hit the bottom wall.
    //drawStyle->style=SoDrawStyle::INVISIBLE;
    initialized =FALSE;
    running=FALSE;
  } else {
    ballTransform->translation.setValue(x, y, z);
  }
}

void Track :: 
do_dynamics_collisions(float seconds, //Time to iterate over.
		       float *o_pos, //ball position.
		       float *o_vel, //ball velocity.
		       float *n_pos, //new ball position.
		       float *n_vel //new ball velocity. 
		       )
{
  int i;
  float new_v;
  int in_goal1 = FALSE;
  int in_goal2 = FALSE;

  //bp_threshold=(BALL_RADIUS+PADDLE_RADIUS)*(BALL_RADIUS + PADDLE_RADIUS);

  for( i = 0; i < 2; i++ )  {
    new_v =  o_vel[i]*(1.0 - friction*seconds);
    n_pos[i] =o_pos[i] + 0.5*(o_vel[i] + new_v)*seconds;
    n_vel[i] = new_v;
  }

  /* Handle ball wall collisions. */
  if ( n_pos[XX] < -board_width_div2 + BALL_RADIUS ) {
    n_pos[XX] = 2*(-board_width_div2 + BALL_RADIUS) - n_pos[XX];
    n_vel[XX] = -restitution*n_vel[XX];
    //g->status_bits |= LEFT_CONTACT;
    //printf("++++ LC:%0.3f\n", n_vel[XX]);
  }
  if ( n_pos[XX] > board_width_div2 - BALL_RADIUS ) {
    n_pos[XX] = 2*(board_width_div2 - BALL_RADIUS) - n_pos[XX];
    n_vel[XX] = -restitution*n_vel[XX];
    //g->status_bits |= RIGHT_CONTACT;
  }
  if ( n_pos[YY] < -board_length_div2 + BALL_RADIUS ) {
    n_pos[YY] = 2*(-board_length_div2 + BALL_RADIUS) - n_pos[YY];
    n_vel[YY] = -restitution*n_vel[YY];
    //g->status_bits |= BOTTOM_CONTACT;
    if ( ( goal2_left <= n_pos[XX] ) &&
	 ( goal2_right >= n_pos[XX] ) )
      in_goal2 = TRUE;
  }
  if ( n_pos[YY] > board_length_div2 - BALL_RADIUS ) {
    n_pos[YY] = 2*(board_length_div2 - BALL_RADIUS) - n_pos[YY];
    n_vel[YY] = -restitution*n_vel[YY];
    //->status_bits |= TOP_CONTACT;
    if ( ( goal1_left <= n_pos[XX] ) &&
	 ( goal1_right >= n_pos[XX] ) )
      in_goal1 = TRUE;
  }
  if(in_goal1==in_goal2) return; //used to eliminate warning.
}


