#ifndef AIR_HOCKEY_H
#define AIR_HOCKEY_H
/*****************************************************************************/
#include "track.hpp"
#include "HockeyData.h"



/*****************************************************************************/

typedef struct sample
{
  int status_bits;
  float time;
  float ball[2];
  float paddle1[2];
  float paddle2[2];
  //float ballV[2];
  //float paddle1V[2];
} Sample;

/*****************************************************************************/

typedef struct communication_buffer
{
  /* game borders */
  float left;
  float right;
  float top;
  float bottom;

  /* goals */
  float goal1_left;
  float goal1_right;
  float goal2_left;
  float goal2_right;

  float ball_radius;
  float paddle1_radius;
  float paddle2_radius;

  float time;

  float ball[3];
  float paddle1[3];
  float paddle2[3];

  int status_bits;

  int score1;
  int score2;

  float max_reach;

  Track * track_ball;

} Communication_Buffer;

/*****************************************************************************/

typedef struct game
{
  /* game borders */
  float left;
  float right;
  float top;
  float bottom;

  /* goals */
  float goal1_left;
  float goal1_right;
  int in_goal1;
  int score1;

  float goal2_left;
  float goal2_right;
  int in_goal2;
  int score2;

  /* constants affecting dynamics */
  float friction;
  float restitution;

  /* states */
  float time;
  float last_time;
  int status_bits;

  /* ball location and velocity */
  float ball[ND];
  float ball_velocity[ND];

  /* paddle1 location and velocity */
  float paddle1[ND];
  float paddle1_velocity[ND];

  /* paddle2 location and velocity */
  float paddle2[ND];
  float paddle2_velocity[ND];

  float ball_radius;
  float paddle1_radius;
  float paddle2_radius;

  float bp_threshold;

  float max_reach;

} Game;
/*****************************************************************************/

void init_game( Communication_Buffer *cb );
void simulate_up_to_time( float stop_time, float mouse_x, float mouse_y,
			 Communication_Buffer *cb );
void read_samples( char *filename );
void replay( Communication_Buffer *cb );

void resetHockeyBoard(int side) ;

#endif
