
#include "togl.h"
#include <math.h>
#include <stdlib.h>
#include <tcl.h>
#include <tk.h>
#include <GL/glu.h>
#include <stdio.h>
#include <GL/gl.h>
#include <iostream.h>

#include "air-hockey.h"
#include "track.hpp"

#include <Inventor/SoDB.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/nodes/SoDirectionalLight.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/nodes/SoTranslation.h>
#include <Inventor/nodes/SoRotation.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoMaterialBinding.h>
#include <Inventor/nodes/SoNormal.h>
#include <Inventor/nodes/SoNormalBinding.h>
#include <Inventor/nodes/SoEventCallback.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/events/SoMouseButtonEvent.h>
#include <Inventor/events/SoLocation2Event.h>
#include <Inventor/events/SoKeyboardEvent.h>
#include <Inventor/sensors/SoTimerSensor.h>
#include <Inventor/sensors/SoAlarmSensor.h>

#ifndef M_PI
#  define M_PI 3.14159265
#endif

SoSeparator *mainRoot;

SbViewportRegion * myViewport;
SoGLRenderAction * myRenderAction;

#define MOUSE_SCALE_FACTOR     0.0009
#define TIME_STEP	       0.001

#define WINDOW_X               500
#define WINDOW_Y               600
#define SCALE_FACTOR 0.8f

extern int use_nn;
extern Game *game;
/*****************************************************************************/

Communication_Buffer the_communication_buffer, *communication_buffer;

SbBool  isMouseDown = FALSE;
SbVec2s mouseLocation;

SoTransform *ballTransform;
SoTransform *paddleTransform;
SoTransform *paddle2Transform;
SoTransform *forearmTransform;
SoTransform *shoulderTransform;
SoTransform *elbowTransform;
SoTransform *upperarmTransform;
SoTransform *pelvis;

SoTransform *uArmRTrans;
SoTransform *lArmRTrans;
SoTransform *handRTrans;
SoTransform *torsoTrans;
SoTransform *headTrans;
SoTransform *uArmLTrans;

SoTimerSensor *timer;
SoAlarmSensor *resetTimer;

float UPPERARM_LENGTH=0.24f;
float FOREARM_LENGTH=0.21f;
float FOREARM_RADIUS=(float) (PADDLE_RADIUS/3.5*2.0);
float SHOULDER_RADIUS=PADDLE_RADIUS/2.0;
float ELBOW_RADIUS=PADDLE_RADIUS/2.0;
float SHOULDER_HEIGHT=0.2f;
float SHOULDER_POSX=-0.06f;
float SHOULDER_POSY=(float)(BOARD_LENGTH/2+0.01);

int time_initialized = 0;
SbTime startTime;
SbTime animationTime;


SoPerspectiveCamera *cam;

//SoXtExaminerViewer *theViewer; 
SoSeparator **bodyNode;
SoTranslation **translationNode;
SoRotation **rotationNode;
SbString *rotationName;
SbString *translationName;
SbString *bodyName;
int *numBodyModels;
SbString **bodyModelName;
SbString *groundModelName;
int numGroundModels;
int numBodies;
SbString creatureName;
char  namelist[ 100 ][ 100 ];

Tcl_Interp * interpGlobal;
SoSeparator *ctrRoot;

typedef float float3[3];
typedef float float4[4];

static int PlayBack=0;

Track * track_ball;
struct Togl *g_togl;

static void reshape( struct Togl *togl);
/************************************************************************/
/************************************************************************/

//int SavePPM(struct Togl *togl, int argc, char *argv[])
int SavePPM()
{
  char *pixels;
  static char *output;
  static int im_num=0;
  int w=Togl_Width(g_togl), h=Togl_Height(g_togl), i, j, k;
  //int w=500, h=600, i, j, k;
  FILE * out;
  char fn[256];  int size = w*h*3;
  output = (char *) malloc(size);
  pixels = (char *) grabPixels(1, w, h);

  sprintf(fn, "../images/sam%d.ppm", im_num++);
  out = fopen(fn, "wb");
  if(!out) {
	  printf("Error -- could not open %s for writing.\n",fn);
	  printf("Ensure the directory exists.\n");
	  exit(-1);
  }
  fprintf(out, "P6\n%d %d\n255\n", w, h);
  k= size-1;
  for(i=0;i<h-1;i++)
    for(j=w-1;j>=0;j--) {
      output[i*w*3+j*3+2] = pixels[k--];
      output[i*w*3+j*3+1] = pixels[k--];
      output[i*w*3+j*3]   = pixels[k--];
    }
  //printf("Saved image number %d!\n", SimViewDisplayData->imageNumber-1);
  fwrite(output, 1, size, out);
  fclose(out);
  free(pixels);
  return(0);
}

/************************************************************************/

SoSeparator *
GetSceneGraph()
{

    // Setup Shoulder.
    SoSeparator   *shoulderRoot        = new SoSeparator;
    SoSeparator   *shoulderCache       = new SoSeparator;
    SoMaterial    *shoulderMaterial    = new SoMaterial;
    SoSphere      *shoulder            = new SoSphere;
    
    shoulderTransform = new SoTransform;

    shoulderTransform->translation.setValue(SHOULDER_POSX,
					    SHOULDER_POSY, SHOULDER_HEIGHT);
    shoulder->radius = SHOULDER_RADIUS;
    shoulderMaterial->shininess = .8f;
    shoulderMaterial->ambientColor.setValue(SbColor(0.8f, 0.1f, 0.0f));
    shoulderMaterial->diffuseColor.setValue(SbColor(1.0f, 0.0f, 0.0f));
    shoulderMaterial->specularColor.setValue(SbColor(1.0f, 1.0f, 1.0f));
    shoulderCache->renderCaching = SoSeparator::ON;
    shoulderCache->addChild(shoulderMaterial);
    shoulderCache->addChild(shoulder);
    shoulderRoot->addChild(shoulderTransform);
    shoulderRoot->addChild(shoulderCache);

    // Setup Elbow.
    SoSeparator   *elbowRoot        = new SoSeparator;
    SoSeparator   *elbowCache       = new SoSeparator;
    SoSphere      *elbow            = new SoSphere;
    
    elbowTransform = new SoTransform;

    elbow->radius = ELBOW_RADIUS;
    elbowCache->renderCaching = SoSeparator::ON;
    elbowCache->addChild(shoulderMaterial);
    elbowCache->addChild(elbow);
    elbowRoot->addChild(elbowTransform);
    elbowRoot->addChild(elbowCache);

    // Setup forearm.
    SoSeparator   *forearmRoot        = new SoSeparator;
    SoSeparator   *forearmCache       = new SoSeparator;
    SoMaterial    *forearmMaterial     = new SoMaterial;
    SoCube    *forearm            = new SoCube;
    
    forearmTransform = new SoTransform;

    forearm->depth = FOREARM_RADIUS;
    forearm->height = FOREARM_RADIUS;
    forearm->width = FOREARM_LENGTH;
    forearmMaterial->shininess = .8f;
    forearmMaterial->ambientColor.setValue(SbColor(1.0, 0.0, 1.0));
    forearmMaterial->diffuseColor.setValue(SbColor(1.0, 0.0, 1.0));
    forearmMaterial->specularColor.setValue(SbColor(1, 1, 1));
    forearmCache->renderCaching = SoSeparator::ON;
    forearmCache->addChild(forearmMaterial);
    forearmCache->addChild(forearm);
    forearmRoot->addChild(forearmTransform);
    forearmRoot->addChild(forearmCache);

    // Setup upperarm.
    SoSeparator   *upperarmRoot        = new SoSeparator;
    SoSeparator   *upperarmCache       = new SoSeparator;
    SoCube        *upperarm            = new SoCube;
    
    upperarmTransform = new SoTransform;

    upperarm->depth = FOREARM_RADIUS;
    upperarm->height = FOREARM_RADIUS;
    upperarm->width = UPPERARM_LENGTH;
    upperarmCache->renderCaching = SoSeparator::ON;
    upperarmCache->addChild(forearmMaterial);
    upperarmCache->addChild(upperarm);
    upperarmRoot->addChild(upperarmTransform);
    upperarmRoot->addChild(upperarmCache);

    // Setup goals.
    SoSeparator   *goal1Root        = new SoSeparator;
    SoSeparator   *goal1Cache       = new SoSeparator;
    SoSeparator   *goal2Root        = new SoSeparator;
    SoSeparator   *goal2Cache       = new SoSeparator;
    SoMaterial    *goalMaterial     = new SoMaterial;
    SoCube        *goal1            = new SoCube;
    SoCube        *goal2            = new SoCube;
    SoTranslation *goal1Translation = new SoTranslation;
    SoTranslation *goal2Translation = new SoTranslation;

    goal1Translation->translation.setValue( 0.0f, (float) (BOARD_LENGTH/2.0), 0.001f);
    goal2Translation->translation.setValue( 0.0f, (float) (-BOARD_LENGTH/2.0), 0.001f );
    goal1->width = BOARD_WIDTH/3;
    goal1->depth = 0.01f;
    goal1->height = 0.001f;
    goal2->width = BOARD_WIDTH/3;
    goal2->depth = 0.01f;
    goal2->height = 0.001f;
    goalMaterial->shininess = .8f;
    goalMaterial->ambientColor.setValue(SbColor(1.0f, 0.0f, 1.0f));
    goalMaterial->diffuseColor.setValue(SbColor(1.0, 0.0, 1.0));
    goalMaterial->specularColor.setValue(SbColor(1, 1, 1));
    goal1Cache->renderCaching = SoSeparator::ON;
    goal1Cache->addChild(goalMaterial);
    goal1Cache->addChild(goal1);
    goal1Root->addChild(goal1Translation);
    goal1Root->addChild(goal1Cache);
    goal2Cache->renderCaching = SoSeparator::ON;
    goal2Cache->addChild(goalMaterial);
    goal2Cache->addChild(goal2);
    goal2Root->addChild(goal2Translation);
    goal2Root->addChild(goal2Cache);

    // Midline.
    SoSeparator   *midlineRoot        = new SoSeparator;
    SoSeparator   *midlineCache       = new SoSeparator;
    SoMaterial    *midlineMaterial     = new SoMaterial;
    SoCube        *midline            = new SoCube;

    midline->width = BOARD_WIDTH;
    midline->depth = 0.01f;
    midline->height = 0.002f;
    midlineMaterial->shininess = .8f;
    midlineMaterial->ambientColor.setValue(SbColor(1.0, 0.0, 1.0));
    midlineMaterial->diffuseColor.setValue(SbColor(1.0, 0.0, 1.0));
    midlineMaterial->specularColor.setValue(SbColor(1, 1, 1));
    midlineCache->renderCaching = SoSeparator::ON;
    midlineCache->addChild(midlineMaterial);
    midlineCache->addChild(midline);
    midlineRoot->addChild(midlineCache);

    // Setup the scene graph for the ball
    SoSeparator   *ballRoot        = new SoSeparator;
    SoSeparator   *ballCache       = new SoSeparator;
    SoMaterial    *ballMaterial    = new SoMaterial;
    SoCylinder      *ball            = new SoCylinder;

    ballTransform = new SoTransform;
    
    ball->radius = BALL_RADIUS;
    ball->height = BALL_HEIGHT;
    ballMaterial->shininess = .8f;
    ballMaterial->ambientColor.setValue(SbColor(1.0, 0.0, 0.0));
    ballMaterial->diffuseColor.setValue(SbColor(1.0, 0.0, 0.0));
    ballMaterial->specularColor.setValue(SbColor(1, 1, 1));
    ballCache->renderCaching = SoSeparator::ON;
    ballCache->addChild(ballMaterial);
    ballCache->addChild(ball);
    ballRoot->addChild(ballTransform);
    ballRoot->addChild(ballCache);

    ballTransform->
      rotation.setValue(SbVec3f(1.0f, 0.0f, 0.0f), (float) (M_PI/2.0) );

    // Setup the scene graph for the paddle
    SoSeparator   *paddleRoot        = new SoSeparator;
    SoSeparator   *paddleCache       = new SoSeparator;
    SoMaterial    *paddleMaterial    = new SoMaterial;
    SoCylinder    *paddle            = new SoCylinder;

    paddleTransform = new SoTransform;
    paddle->radius = PADDLE_RADIUS;
    paddle->height = PADDLE_HEIGHT;
    paddleMaterial->shininess = .8f;
    paddleMaterial->ambientColor.setValue(SbColor(0.0, 1.0, 0.0));
    paddleMaterial->diffuseColor.setValue(SbColor(0.0, 1.0, 0.0));
    paddleMaterial->specularColor.setValue(SbColor(1, 1, 1));
    paddleCache->renderCaching = SoSeparator::ON;
    paddleCache->addChild(paddleMaterial);
    paddleCache->addChild(paddle);
    paddleRoot->addChild(paddleTransform);
    paddleRoot->addChild(paddleCache);

    // Setup the scene graph for paddle2
    SoSeparator   *paddle2Root        = new SoSeparator;
    SoSeparator   *paddle2Cache       = new SoSeparator;
    SoMaterial    *paddle2Material    = new SoMaterial;
    SoCylinder    *paddle2            = new SoCylinder;

    paddle2Transform = new SoTransform;
    paddle2->radius = PADDLE_RADIUS;
    paddle2->height = PADDLE_HEIGHT;
    paddle2Material->shininess = .8f;
    paddle2Material->ambientColor.setValue(SbColor(0.0, 0.0, 1.0));
    paddle2Material->diffuseColor.setValue(SbColor(0.0, 0.0, 1.0));
    paddle2Material->specularColor.setValue(SbColor(1, 1, 1));
    paddle2Cache->renderCaching = SoSeparator::ON;
    paddle2Cache->addChild(paddle2Material);
    paddle2Cache->addChild(paddle2);
    paddle2Root->addChild(paddle2Transform);
    paddle2Root->addChild(paddle2Cache);

    paddleTransform->
      rotation.setValue(SbVec3f(1.0, 0.0, 0.0), (float) (M_PI/2.0) );
    paddle2Transform->
      rotation.setValue(SbVec3f(1.0, 0.0, 0.0), (float) (M_PI/2.0) );
    // Generate a scene graph representing the maze specified by the array
    SoSeparator *mazeRoot          = new SoSeparator;
    SoSeparator *mazeCache         = new SoSeparator;
    SoMaterial  *floorMaterial     = new SoMaterial;
    SoCube      *floor             = new SoCube;
    SoMaterial  *mazeMaterial      = new SoMaterial;
    SoMaterialBinding *mazeMtlBind = new SoMaterialBinding;
    SoNormal    *mazeNorms         = new SoNormal;
    SoNormalBinding *mazeNormBind  = new SoNormalBinding;

    // Create walls for the boundary of the maze
    mazeNorms->vector.set1Value(0, SbVec3f( 0.0, -1.0,  0.0));    
    mazeNorms->vector.set1Value(1, SbVec3f( 0.0,  1.0,  0.0));    
    mazeNorms->vector.set1Value(2, SbVec3f( 0.0,  0.0,  1.0));    
    mazeNorms->vector.set1Value(3, SbVec3f( 1.0,  0.0,  0.0));    
    mazeNorms->vector.set1Value(4, SbVec3f( 0.0,  0.0, -1.0));    
    mazeNorms->vector.set1Value(5, SbVec3f(-1.0,  0.0,  0.0));
    mazeMaterial->diffuseColor.setValue(SbColor(.8f, .6f, .1f));
    mazeMtlBind->value = SoMaterialBinding::OVERALL;
    mazeNormBind->value = SoNormalBinding::PER_FACE_INDEXED;

    floorMaterial->diffuseColor.setValue(SbColor(.95f, .8f, .1f));
    floorMaterial->shininess = 0.2f;
    floor->width = BOARD_WIDTH;
    floor->depth = 0.001f;
    floor->height = BOARD_LENGTH;

    mazeCache->renderCaching = SoSeparator::ON;
    mazeCache->addChild(floorMaterial);
    mazeCache->addChild(floor);
    mazeCache->addChild(mazeMaterial);
    mazeCache->addChild(mazeNorms);
    mazeCache->addChild(mazeMtlBind);
    mazeCache->addChild(mazeNormBind);

    mazeRoot->addChild(ballRoot);
    mazeRoot->addChild(paddleRoot);
    mazeRoot->addChild(paddle2Root);
    mazeRoot->addChild(goal1Root);
    mazeRoot->addChild(goal2Root);

    mazeRoot->addChild(forearmRoot);
    mazeRoot->addChild(upperarmRoot);
    mazeRoot->addChild(shoulderRoot);
    mazeRoot->addChild(elbowRoot);
    mazeRoot->addChild(midlineRoot);
    mazeRoot->addChild(mazeCache);
	

    return mazeRoot;
}

/*****************************************************************************/
void get_quaternion(SbVec3f v1,  SbVec3f v2, float * q)
{
  int i;
  SbVec3f r0, r1, r2;
  float length;

  r0=v1;
  r1=v2;

  r0.normalize();
  r1.normalize();
  r2= r0.cross(r1);
  r2.normalize();
  r1=r2.cross(r0);

  q[3] = 0.5*sqrt( (double) 1.0 + r0[0] + r1[1] + r2[2] );
  if ( fabs( (double) q[3] ) > 1e-10 )
    {
      q[0] = -0.25*(r2[1] - r1[2])/q[3];
      q[1] = -0.25*(r0[2] - r2[0])/q[3];
      q[2] = -0.25*(r1[0] - r0[1])/q[3];
    }
  else
    {
      q[0] = -0.25*(r2[1] - r1[2]);
      q[1] = -0.25*(r0[2] - r2[0]);
      q[2] = -0.25*(r1[0] - r0[1]);
      length = q[0]*q[0] + q[1]*q[1] + q[2]*q[2] + q[3]*q[3];
      length = sqrt( (double) length );
      for( i = 0; i < 4; i++ )
	q[i] /= length;
    }
}




static SoSeparator *buildCreature() 
{ 
  SoInput in;
  SoSeparator *cr_root;
  FILE *filePtr = NULL;
    
  //fprintf(stderr, "Reading file(s)...\n");
  cout << "Reading file(s)..." << endl;
  filePtr = fopen("corps.iv", "r");
  if (filePtr == NULL) {
    fprintf(stderr, "Error opening file corps.iv\n");
    return (NULL);
  }

  in.setFilePointer(filePtr);
  cr_root = SoDB::readAll(&in);
  if (cr_root == NULL) 
    fprintf(stderr, "Error reading file corps.iv\n");

  fclose(filePtr);
  
  return (cr_root);
}



static int resetGame(struct Togl *togl, int argc, char *argv[])
{
	SoSensor * mySensor=NULL;
	void * vptr=NULL;
	//mouseLocation.setValue(0, 0);

	// Place the puck in the center of the board moving in the correct direction.
	resetHockeyBoard(2);

	// Reset scores.
	communication_buffer->score1=0;
	communication_buffer->score2=0;


	cerr <<"+++++++Reseting game"<<endl;

  upperarmTransform->
	translation.setValue(SHOULDER_POSX-UPPERARM_LENGTH/2, 
			   SHOULDER_POSY, SHOULDER_HEIGHT);

  uArmLTrans->translation.setValue( 0.21f,
					(float)(BOARD_LENGTH/2.0+0.05),
					0.07f);
  lArmRTrans->translation.setValue( -0.1f,
					(float)(BOARD_LENGTH/2.0+0.05-0.2),
					0.2f);
  lArmRTrans->translation.setValue( 0.0f,
					(float)(BOARD_LENGTH/2.0+0.05),
					0.3f);
  lArmRTrans->translation.setValue( 0.0,  0.0,	0.0);
  forearmTransform->
    translation.setValue(0, 0, 0);


  torsoTrans->translation.setValue( 0.075f,
				    (float) (BOARD_LENGTH/2+0.05),
				    0.205f);
  headTrans->translation.setValue( 0.075f,
				   (float) (BOARD_LENGTH/2.0-0.05),
				   0.3f);
  headTrans->rotation.setValue(SbVec3f(1.0f, 0.0f, 0.0f), 3.8f);

  headTrans->
    scaleFactor.setValue(SCALE_FACTOR, SCALE_FACTOR, SCALE_FACTOR);
  uArmLTrans->
    scaleFactor.setValue(SCALE_FACTOR, SCALE_FACTOR, SCALE_FACTOR);
  uArmRTrans->
    scaleFactor.setValue(SCALE_FACTOR, SCALE_FACTOR, SCALE_FACTOR);
  lArmRTrans->
    scaleFactor.setValue(SCALE_FACTOR, SCALE_FACTOR, SCALE_FACTOR);
  handRTrans->
    scaleFactor.setValue(SCALE_FACTOR, SCALE_FACTOR, SCALE_FACTOR);
  torsoTrans->
    scaleFactor.setValue(SCALE_FACTOR, SCALE_FACTOR, SCALE_FACTOR);

  //RunSimulation( vptr,  mySensor);

   //reshape(togl);
  return 0;
    
}


////////////////////////////////////////////////////////////////////////
//
// Description:
//    Callback routine for receiving keyboard events.
//
// Use: private

static void
ReceiveKeyboardEvent( void *, SoEventCallback *cb )

//////////////////////////////////////////////////////////////////////////
{
    if (SO_KEY_PRESS_EVENT(cb->getEvent(), R)) {
        resetGame(NULL, NULL, NULL);
        cb->setHandled(); 
    } 
    if (SO_KEY_PRESS_EVENT(cb->getEvent(), NUMBER_1)) {
        resetGame(NULL, NULL, NULL);
        cb->setHandled(); 
    }
    if (SO_KEY_PRESS_EVENT(cb->getEvent(), T)) {
      track_ball->Toggle();
    }
    if (SO_KEY_PRESS_EVENT(cb->getEvent(), N)) {
      if(use_nn) {
		use_nn=0;
		cout << "Using Linear Regression"<<endl;
      } else {
		cout << "Using Neural Net"<<endl;
		use_nn=1;
      }
    }
}


////////////////////////////////////////////////////////////////////////
//
// Description:
//    Callback routine for receiving mouse button press.
//    Schedule the timer sensor on mouse down.
//
// Use: private

static void
ReceiveMouseButtonEvent( void *, SoEventCallback *cb )

//
////////////////////////////////////////////////////////////////////////
{
    if (SO_MOUSE_PRESS_EVENT(cb->getEvent(), BUTTON1)) {

        const SbVec2s &pos = cb->getEvent()->getPosition();
        mouseLocation.setValue(pos[0], pos[1]);

        if (!timer->isScheduled()) {
            animationTime.setValue( SbTime::getTimeOfDay().getValue() );
            timer->schedule();
        }

        isMouseDown = TRUE;
        cb->setHandled();
    }
    else if (SO_MOUSE_RELEASE_EVENT(cb->getEvent(), BUTTON1)) {

        isMouseDown = FALSE;
        cb->setHandled();
    }
}

/************************************************************************/

////////////////////////////////////////////////////////////////////////
//
// Description:
//    Callback routine for receiving mouse motion
//
// Use: private

static void
ReceiveMouseMotionEvent( void *, SoEventCallback *cb )

//
////////////////////////////////////////////////////////////////////////
{
	/*
    if (!isMouseDown)
        return;
	*/

    const SbVec2s &pos = cb->getEvent()->getPosition();
    mouseLocation.setValue((short)pos[0], (short)pos[1]);
    cb->setHandled();
}


////////////////////////////////////////////////////////////////////////
//
// Description:
//    Callback routine for resetting the game back to its initial state
//
// Use: private

static void
resetCB( void *, SoSensor * )
{
    resetGame(NULL, NULL, NULL);
}


static GLuint limit;
static GLuint count = 1;

static GLubyte polycolor[4] = {255, 255, 255, 255};

static void draw( struct Togl *togl )
{
	static int firsttime = 1;
   glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

   myRenderAction->apply(mainRoot);

	if(firsttime) {
		firsttime = 0;
		reshape(togl);
	}

   Togl_SwapBuffers(togl);

}



static void idle( struct Togl *togl )
{

    float mx, my;
    static float seconds=0;
	static int score1=0,score2=0;
    double d, dx, dy, dz, a;
    double r, theta, theta_z;
    double elbow_x, elbow_y;
    double elbow_z=SHOULDER_HEIGHT/2;
    double mid_x, mid_y, mid_z;
    float upper_q[4], lower_q[4], hand_q[4];
	static int count=0, dataNum=0;

    SbTime TimeFromStart;
    SbVec3f sh, elbow, wrist, fingers, v1, v2;

	if ( !time_initialized ) {
		startTime.setValue( SbTime::getTimeOfDay().getValue() );
		time_initialized = 1;
	}
	
	if(PlayBack) {  //Playing back data, set objects from the data file.
		static Sample samples[NPOINTS];
		static Sample *sample_ptr=samples;
		static int num_samples=1000;  //the number of samples to be read and playedback.

		if(sample_ptr==samples) {  //Initialize the samples.
			FILE *stream;
			char filename[256];
			sprintf(filename,"%s", "d00038");
			
			stream = fopen( filename, "r" );
			if ( stream == NULL )
			{
				fprintf( stderr, "Can't open %s for read.\n", filename );
				exit( -1 );
			}
			printf( "Reading data from %s\n", filename );
			for( sample_ptr = samples; sample_ptr - samples < num_samples; sample_ptr++ )
			{
				fscanf( stream, "%d %g %g %g %g %g %g %g\n",
					&(sample_ptr->status_bits),
					&(sample_ptr->time),
					&(sample_ptr->ball[0]),
					&(sample_ptr->ball[1]),
					&(sample_ptr->paddle1[0]),
					&(sample_ptr->paddle1[1]),
					&(sample_ptr->paddle2[0]),
					&(sample_ptr->paddle2[1]) );
			}
			fclose( stream );
			printf( "done.\n" );
		}
		
		if ( sample_ptr - samples >= num_samples ) {
			sample_ptr = samples;
			sample_ptr++;
		}
		
		communication_buffer->time = sample_ptr->time;
		communication_buffer->ball[0] = sample_ptr->ball[0];
		communication_buffer->ball[1] = sample_ptr->ball[1];
		communication_buffer->paddle1[0] = sample_ptr->paddle1[0];
		communication_buffer->paddle1[1] = sample_ptr->paddle1[1];
		communication_buffer->paddle2[0] = sample_ptr->paddle2[0];
		communication_buffer->paddle2[1] = sample_ptr->paddle2[1]; 
		communication_buffer->status_bits = sample_ptr->status_bits;
		
		sample_ptr++;
	} else {
		animationTime.setValue( SbTime::getTimeOfDay().getValue() );

	    TimeFromStart.setValue( (animationTime - startTime).getValue() );
		seconds = TimeFromStart.getValue();
	    //seconds+=0.03f;
	    mx = (float)(mouseLocation[0] - Togl_Width(togl)/2)*0.0012;
		my = -(float)(mouseLocation[1] - Togl_Height(togl)/2)*0.0031;


		// Set the status bit to only the RESET_PRESSED bit if it was set.
		game->status_bits &= RESET_PRESSED;

		// Find out if the mouse position is in a location that would show that the
		// human is playing.
		if(mx<-0.29 || mx>0.29 || my<-0.8 || my>0.0) {
			game->status_bits |= MOUSE_NOT_IN_WINDOW;
			//cout <<"---Outside Window.------"<<endl;
		}

	    simulate_up_to_time( seconds, mx, my, communication_buffer );
	}
	    //Compute elbow position.
    wrist[0]=communication_buffer->paddle2[0]-0.07;
    wrist[1]=communication_buffer->paddle2[1]+0.00+
      (0.4-communication_buffer->paddle2[1])*0.2;
    wrist[2]=0.055f;

    fingers[0]=communication_buffer->paddle2[0]+.01;
    fingers[1]=communication_buffer->paddle2[1];
    fingers[2]=0.055f;

    dx=SHOULDER_POSX-wrist[0];
    dy=SHOULDER_POSY-wrist[1]; 
    dz=SHOULDER_HEIGHT-wrist[2];

    //Distance from shoulder to wrist.
    d=sqrt(dx*dx+dy*dy+dz*dz);

    //Distance to point on line that is at the elbow z position.
    a=(square(UPPERARM_LENGTH)+square(d)-square(FOREARM_LENGTH))/(2*d);
    theta_z=atan((SHOULDER_POSY-wrist[1])/(SHOULDER_HEIGHT-wrist[2]));

    r=sqrt(square(UPPERARM_LENGTH) - square(a));
    
    theta=atan2(dx,dy);
    mid_x=SHOULDER_POSX-a*sin(theta_z)*sin(theta);
    mid_y=SHOULDER_POSY-a*sin(theta_z)*cos(theta);
    elbow_x=mid_x-r*cos(theta);
    elbow_y=mid_y+r*sin(theta);
    mid_z=SHOULDER_HEIGHT-a*cos(theta_z);
    elbow_z=mid_z;

    sh[0]=SHOULDER_POSX;
    sh[1]=SHOULDER_POSY;
    sh[2]=SHOULDER_HEIGHT;
    elbow[0]=elbow_x;
    elbow[1]=elbow_y;
    elbow[2]=elbow_z;

	v1.setValue(1.0, 1.0, 0.0);
    get_quaternion( (elbow-sh), (wrist-elbow), upper_q);
    get_quaternion( (wrist-elbow), (sh-elbow), lower_q);
    get_quaternion( (fingers-wrist), v1, hand_q);
    
	if(communication_buffer->ball[0] == 0.0f && communication_buffer->ball[1] == 0.0f) {
		cerr<<"****ALL ZEROS****"<<endl;
	}
	ballTransform->translation.setValue( communication_buffer->ball[0],
			communication_buffer->ball[1],
			BALL_HEIGHT/2);
    paddleTransform->translation.setValue( communication_buffer->paddle1[0],
					   communication_buffer->paddle1[1],
					   PADDLE_HEIGHT/2);

    paddle2Transform->translation.setValue( communication_buffer->paddle2[0],
					    communication_buffer->paddle2[1],
					    PADDLE_HEIGHT/2);

	
    //Put hand into position.
    handRTrans->translation.setValue( 0.5*(wrist[0]+fingers[0]),
    			      0.5*(wrist[1]+fingers[1]),
    			      0.5*(wrist[2]+fingers[2]));
    
	mid_x = (elbow_x+wrist[0])/2;
	mid_y = (elbow_y+wrist[1])/2;
    mid_z= (elbow_z+wrist[2])/2;
    headTrans->rotation.setValue(SbVec3f(0.0, 0.0, 1.0), -theta);
    
    forearmTransform->
      translation.setValue(mid_x,  mid_y,  mid_z);
    lArmRTrans->
      translation.setValue(mid_x,   mid_y,  mid_z);

    mid_x = (elbow_x+SHOULDER_POSX)/2;
    mid_y = (elbow_y+SHOULDER_POSY)/2;
    mid_z= (elbow_z+SHOULDER_HEIGHT)/2;

    upperarmTransform->
      translation.setValue(mid_x, mid_y,  mid_z);
    uArmRTrans->
      translation.setValue(mid_x, mid_y,  mid_z);

    elbowTransform->
      translation.setValue(elbow_x, 
			   elbow_y,
			   elbow_z);

    upperarmTransform->
    rotation.setValue(upper_q);    
    forearmTransform->
      rotation.setValue(lower_q);
    lArmRTrans->
      rotation.setValue(lower_q);

    handRTrans->
      rotation.setValue(hand_q);
    uArmRTrans->
      rotation.setValue(upper_q);

	// If the scores have changed, update the scores in the tcl window and place the ball in the middle of the board.
	if( score1 != communication_buffer->score1 || score2 != communication_buffer->score2 ) {
		
		Tcl_Interp *interp = Togl_Interp(togl);
		char updateString[256];
		int rv;

		if(score1 != communication_buffer->score1) { // Human Scored.
			resetHockeyBoard(1);
		} else {  // Cyber player scored.
			resetHockeyBoard(2);
		}

		score1 = communication_buffer->score1;
		score2 = communication_buffer->score2;
		
		sprintf(updateString, "GUIUpdate %d %d", communication_buffer->score1,
			communication_buffer->score2);
		rv = Tcl_GlobalEval(interp, updateString);
		if (TCL_OK != rv) {
			fprintf(stderr, "Callback failed.  %s\n", interp->result);
			fflush(stderr);
			exit(-1);
		}

	}

  game->status_bits = 0; // Clear the status bits.
  if(PlayBack)
	SavePPM();
   Togl_PostRedisplay(togl);
}


/* new window size or exposure */
static void reshape( struct Togl *togl)
{
   int width, height ;
   width = Togl_Width(togl);
   height = Togl_Height(togl);
   glViewport(0, 0, (GLint)width, (GLint)height);
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   //cout << "W:"<<width<<" h:"<<height<<endl;
   if (width>height) {
      GLfloat w = (GLfloat) width / (GLfloat) height;
      glFrustum( -w, w, -1.0, 1.0, 0.0, 60.0 );
   }
   else {
      GLfloat h = (GLfloat) height / (GLfloat) width;
      glFrustum( -1.0, 1.0, -h, h, 0.0, 60.0 );
   }

   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

}


static void init( struct Togl *togl )
{

   static GLfloat red[4] = {0.8f, 0.1f, 0.0f, 1.0f };
   static GLfloat green[4] = {0.0f, 0.8f, 0.2f, 1.0f };
   static GLfloat blue[4] = {0.2f, 0.2f, 1.0f, 1.0f };
   static GLfloat pos[4] = {5.0f, 5.0f, 10.0f, 0.0f };
   int width, height ;

   glLightfv( GL_LIGHT0, GL_POSITION, pos );
   glEnable( GL_CULL_FACE );
   glEnable( GL_LIGHTING );
   glEnable( GL_LIGHT0 );
   glEnable( GL_DEPTH_TEST );

   glEnable( GL_NORMALIZE );
	g_togl = togl;
	
   width = Togl_Width(togl);
   height = Togl_Height(togl);
   glViewport(0, 0, (GLint)width, (GLint)height);
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   //cout << "W:"<<width<<" h:"<<height<<endl;
   if (width>height) {
      GLfloat w = (GLfloat) width / (GLfloat) height;
      glFrustum( -w, w, -1.0, 1.0, 0.0, 60.0 );
   }
   else {
      GLfloat h = (GLfloat) height / (GLfloat) width;
      glFrustum( -1.0, 1.0, -h, h, 0.0, 60.0 );
   }

   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
}


int movePaddle( struct Togl *togl, int argc, char *argv[] )
{
   Tcl_Interp *interp = Togl_Interp(togl);

   /* error checking */
   if (argc != 4) {
      Tcl_SetResult( interp,
                     "wrong # args: should be \"pathName xrot yrot\"",
                     TCL_STATIC );
      return TCL_ERROR;
   }

   mouseLocation[0] =atof( argv[2] );
   mouseLocation[1] =atof( argv[3] );
   Togl_PostRedisplay(togl);

   /* Let result string equal value */
   strcpy( interp->result, argv[2] );
   return TCL_OK;
}


//
// Take care of the radio button press.  This commandis called instead
// polling to increase simulation speed.
int handleRadios(struct Togl *togl, int argc, char *argv[])
{
   Tcl_Interp *interp = Togl_Interp(togl);

   /* error checking */
   if (argc != 4) {
      Tcl_SetResult( interp,
                     "wrong # args: should be \"pathName trackBall algorithm\"",
                     TCL_STATIC );
      return TCL_ERROR;
   }

   if(*argv[2] == '1') { //Show the trackball
	   if(!track_ball->GetShow())
		   track_ball->Toggle();
   } else { //Hide the trackball
	   if(track_ball->GetShow())
		   track_ball->Toggle();
   }

   if(*argv[3] == '1') { //use neural net
	   use_nn = 1;
   } else { //use linear regression.
	   use_nn=0;
   }

   /* Let result string equal value */
   strcpy( interp->result, argv[2] );
   return TCL_OK;
}

int my_init(Tcl_Interp *interp) 
{
   /*
    * Initialize Tcl, Tk, and the Togl widget module.
    */

   if (Tcl_Init(interp) == TCL_ERROR) {
      return TCL_ERROR;
   }
   if (Tk_Init(interp) == TCL_ERROR) {
      return TCL_ERROR;
   } 
   if (Togl_Init(interp) == TCL_ERROR) {
      return TCL_ERROR;
   }

   /*
    * Specify the C callback functions for widget creation, display,
    * and reshape.
    */
    SoSeparator *SceneRoot;

    communication_buffer = &the_communication_buffer;

   
    cout << endl<<"Air-Hockey Simulator"<<endl<<endl;
	/*
    cout << "While playing, the follow keys are active:"<<endl;
    cout << "T - Toggles the tack ball, a ball that follows a tratectory"<<endl;
    cout << "    that the ball should follow."<<endl;
    cout << "N - Toggles between the use of the neural-network or the"<<endl;
    cout << "    local-regression function to decide how to make a shot."<<endl;
    cout << endl;
	*/

    /* Initialize CGA game stuff. */
    communication_buffer->max_reach=
      sqrt( (square( (FOREARM_LENGTH+UPPERARM_LENGTH)))-
	    (square(SHOULDER_HEIGHT)) );
    cout << "max:" << communication_buffer->max_reach <<endl; 
    init_game( communication_buffer );

    // Initialize Inventor
    SoDB::setDelaySensorTimeout(SbTime(0.0));

	SoDB::init();

    SceneRoot = GetSceneGraph();
	
    // Setup event callbacks
    SoEventCallback *eventCB = new SoEventCallback;
    eventCB->addEventCallback(SoMouseButtonEvent::getClassTypeId(),
            ReceiveMouseButtonEvent, (void *)&mouseLocation);
    eventCB->addEventCallback(SoLocation2Event::getClassTypeId(),
            ReceiveMouseMotionEvent, (void *)&mouseLocation);
    eventCB->addEventCallback(SoKeyboardEvent::getClassTypeId(),
            ReceiveKeyboardEvent, NULL);

    SbVec3f pos(0.0f, -1.2f, 0.48f);
    SbVec3f lookto(0.0f, 0.1f, 0.0f);

    // Setup the camera so the game is comfortably viewed
    cam = new SoPerspectiveCamera;
    cam->position.setValue(pos);
    cam->orientation.setValue(SbRotation(SbVec3f(0.0, 0.0, -1.0),
                                         lookto-pos));
    cam->focalDistance = (lookto-pos).length();

	cam->aspectRatio=(float)500.0/600.0f;
	cam->heightAngle =(float)M_PI/4.5f;
	cam->nearDistance.setValue(0.03f);
	cam->farDistance.setValue(10.01f);

	mainRoot = new SoSeparator;
    mainRoot->ref();

    //mainRoot->addChild(eventCB);
    mainRoot->addChild(cam);
    mainRoot->addChild(SceneRoot);

    track_ball = new Track(mainRoot);
    if(track_ball == NULL)
      cerr << "No Track Ball"<<endl;
    communication_buffer->track_ball = track_ball;

    ctrRoot = buildCreature(); 
	
    mainRoot->addChild(ctrRoot);

	lArmRTrans=(SoTransform *)SoNode::getByName("R_low_arm_t");
    if(lArmRTrans==NULL) 
      fprintf(stderr,"lArmRTrans NULL.\n");
    uArmLTrans=(SoTransform *)SoNode::getByName("L_up_arm_t");
    if(uArmLTrans==NULL) 
      fprintf(stderr,"lArmLTrans NULL.\n");
    uArmRTrans=(SoTransform *)SoNode::getByName("R_up_arm_t");
    if(uArmRTrans==NULL) 
      fprintf(stderr,"uArmRTrans NULL.\n");
    handRTrans=(SoTransform *)SoNode::getByName("R_hand_t");
    if(handRTrans==NULL) 
      fprintf(stderr,"handRTrans NULL.\n");
    torsoTrans=(SoTransform *)SoNode::getByName("up_torso_t");
    if(torsoTrans==NULL) 
      fprintf(stderr,"torsoTrans NULL.\n");

    headTrans=(SoTransform *)SoNode::getByName("T_Head_t");
    if(headTrans==NULL) 
      fprintf(stderr,"headTrans NULL.\n");

    resetGame(NULL, NULL, NULL);

    myViewport = new SbViewportRegion(WINDOW_X, WINDOW_Y);
    myRenderAction =new SoGLRenderAction(*myViewport);

   Togl_CreateFunc( init );
   //Togl_DestroyFunc( zap );
   Togl_DisplayFunc( draw );
   Togl_ReshapeFunc( reshape );
   Togl_TimerFunc(  idle );
   Togl_CreateCommand("movePaddle",movePaddle);
   Togl_CreateCommand("reset",resetGame);
   Togl_CreateCommand("handleRadios",handleRadios);
   Tcl_SetVar( interp, "tcl_rcFileName", "./hockey.tcl", TCL_GLOBAL_ONLY );
   interpGlobal=interp;
   
   return TCL_OK;
}

int main( int argc, char *argv[] )
{
  Tk_Main(argc,argv,my_init );
  return 0;
}
